#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.common import noerr
from xpra.util.objects import typedict
from xpra.codecs.image import ImageWrapper
from xpra.codecs.pillow.encoder import (
    get_encodings, encode,
    get_version, get_type, get_info,
    selftest,
)


class TestPillow(unittest.TestCase):

    def test_selftest(self):
        for full in (False, True):
            selftest(full)

    def test_module_functions(self):
        numver = tuple(noerr(int, x) or 0 for x in get_version().split("."))
        assert numver > (5, 0)
        assert get_type() == "pillow"
        assert get_info()

    def test_resize(self):
        W = 128
        H = 128
        for speed in (10, 30, 90, 100):
            pixel_data = b"0"*128*128*4
            Bpp = 4
            image = ImageWrapper(0, 0, W, H, pixel_data, "BGRX", 32,
                                 W*Bpp, Bpp, planes=ImageWrapper.PACKED,
                                 thread_safe=True)
            comp = encode("png", image, typedict({
                "speed" : speed,
                "scaled-width" : 64,
                "scaled-height" : 64,
            }))
            assert comp

    def test_grayscale(self):
        W = 128
        H = 128
        pixel_data = b"0"*128*128*4
        Bpp = 4
        image = ImageWrapper(0, 0, W, H, pixel_data, "BGRA", 32,
                             W*Bpp, Bpp, planes=ImageWrapper.PACKED,
                             thread_safe=True)
        comp = encode("png", image, typedict({"grayscale": True}))
        assert comp

    def test_encode_image_formats(self):
        for pixel_format, bpp in {
            "r210": 32,
            "BGRA": 32,
            "BGR565": 16,
            "RLE8": 8,
        }.items():
            for encoding in get_encodings():
                if encoding=="jpeg" and pixel_format!="BGRA":
                    continue
                self.do_test_pixel_format(pixel_format, bpp, encoding)

    def do_test_pixel_format(self, pixel_format, bpp, encoding):
        maxsize = 1024*1024*4
        buf = bytearray(maxsize)
        palette = [(10, 255, 127), (0, 0, 0), (255, 255, 255)] + [(0, 0, 0)]*253
        for transparency in (True, False):
            for quality in (0, 1, 50, 99, 100):
                for speed in (0, 1, 50, 99, 100):
                    for width in (1, 10, 256):
                        for height in (1, 10, 256):
                            size = width*height*bpp
                            pixel_data = bytes(buf[:size])
                            Bpp = bpp//8
                            image = ImageWrapper(0, 0, width, height, pixel_data, pixel_format, 32,
                                                 width*Bpp, Bpp, planes=ImageWrapper.PACKED,
                                                 thread_safe=True, palette=palette)
                            options = {
                                "quality" : quality,
                                "speed" : speed,
                                "alpha" : transparency,
                            }
                            try:
                                comp = encode(encoding, image, typedict(options))
                                if not comp:
                                    raise ValueError(f"{encode} returned {comp}")
                            except Exception as e:
                                raise RuntimeError(f"failed to encode {image} as {encoding} with options {options}") \
                                    from e

    def test_invalid_pixel_format(self):
        width = 32
        height = 32
        bpp = 4
        pixel_format = "invalid"
        pixel_data = bytes(b"0"*bpp*width*height)
        Bpp = bpp//8
        image = ImageWrapper(0, 0, width, height, pixel_data, pixel_format, 32,
                             width*Bpp, Bpp, planes=ImageWrapper.PACKED,
                             thread_safe=True, palette=None)
        try:
            encode("png", image, typedict())
        except Exception:
            pass
        else:
            raise Exception("should not be able to process '%s'" % pixel_format)

    def test_invalid_encoding(self):
        width = 32
        height = 32
        bpp = 4
        pixel_format = "BGRA"
        pixel_data = bytes(b"0"*bpp*width*height)
        Bpp = bpp//8
        image = ImageWrapper(0, 0, width, height, pixel_data, pixel_format, 32,
                             width*Bpp, Bpp, planes=ImageWrapper.PACKED,
                             thread_safe=True, palette=None)
        for encoding in (None, "", True, "hello", 1):
            try:
                encode(encoding, image, typedict())
            except Exception:
                pass
            else:
                raise Exception("'%s' is an invalid encoding" % encoding)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from time import monotonic
from typing import Any, Dict, Tuple
from collections.abc import Sequence

from libc.stdint cimport uint8_t, uint32_t, uintptr_t    # pylint: disable=syntax-error
from libc.stdlib cimport free    # pylint: disable=syntax-error
from libc.string cimport memset  # pylint: disable=syntax-error
from xpra.buffers.membuf cimport buffer_context

from xpra.codecs.image import ImageWrapper
from xpra.codecs.constants import VideoSpec
from xpra.net.compression import Compressed
from xpra.codecs.debug import may_save_image
from xpra.util.env import envbool, envint
from xpra.util.objects import typedict
from xpra.log import Logger
log = Logger("encoder", "webp")


cdef int LOG_CONFIG = envbool("XPRA_WEBP_LOG_CONFIG", False)
cdef int WEBP_THREADING = envbool("XPRA_WEBP_THREADING", True)
cdef int SUBSAMPLING_THRESHOLD = envint("XPRA_WEBP_SUBSAMPLING_THRESHOLD", 80)


cdef inline int MIN(int a, int b):
    if a<=b:
        return a
    return b

cdef inline int MAX(int a, int b):
    if a>=b:
        return a
    return b


cdef extern from *:
    ctypedef unsigned long size_t


DEF WEBP_MAX_DIMENSION = 16383

cdef extern from "webp/encode.h":

    int WebPGetEncoderVersion()

    ctypedef int WebPImageHint
    # WebPImageHint:
    WebPImageHint WEBP_HINT_DEFAULT     #default preset
    WebPImageHint WEBP_HINT_PICTURE     #digital picture, like portrait, inner shot
    WebPImageHint WEBP_HINT_PHOTO       #outdoor photograph, with natural lighting
    WebPImageHint WEBP_HINT_GRAPH       #Discrete tone image (graph, map-tile etc).

    ctypedef int WebPPreset
    WebPPreset WEBP_PRESET_DEFAULT      #default preset.
    WebPPreset WEBP_PRESET_PICTURE      #digital picture, like portrait, inner shot
    WebPPreset WEBP_PRESET_PHOTO        #outdoor photograph, with natural lighting
    WebPPreset WEBP_PRESET_DRAWING      #hand or line drawing, with high-contrast details
    WebPPreset WEBP_PRESET_ICON         #small-sized colorful images
    WebPPreset WEBP_PRESET_TEXT         #text-like

    ctypedef enum WebPEncCSP:
        #chroma sampling
        WEBP_YUV420                     #4:2:0
        WEBP_YUV420A                    #4:2:0 + alpha
        WEBP_CSP_UV_MASK                #bit-mask to get the UV sampling factors
        #WEBP_CSP_ALPHA_BIT              #bit that is set if alpha is present

    ctypedef int WebPEncodingError
    WebPEncodingError VP8_ENC_OK
    WebPEncodingError VP8_ENC_ERROR_OUT_OF_MEMORY
    WebPEncodingError VP8_ENC_ERROR_BITSTREAM_OUT_OF_MEMORY
    WebPEncodingError VP8_ENC_ERROR_NULL_PARAMETER
    WebPEncodingError VP8_ENC_ERROR_INVALID_CONFIGURATION
    WebPEncodingError VP8_ENC_ERROR_BAD_DIMENSION
    WebPEncodingError VP8_ENC_ERROR_PARTITION0_OVERFLOW
    WebPEncodingError VP8_ENC_ERROR_PARTITION_OVERFLOW
    WebPEncodingError VP8_ENC_ERROR_BAD_WRITE
    WebPEncodingError VP8_ENC_ERROR_FILE_TOO_BIG
    WebPEncodingError VP8_ENC_ERROR_USER_ABORT
    WebPEncodingError VP8_ENC_ERROR_LAST

    ctypedef struct WebPConfig:
        int lossless                    #Lossless encoding (0=lossy(default), 1=lossless).
        float quality                   #between 0 (smallest file) and 100 (biggest)
        int method                      #quality/speed trade-off (0=fast, 6=slower-better)

        WebPImageHint image_hint        #Hint for image type (lossless only for now).

        #Parameters related to lossy compression only:
        int target_size                 #if non-zero, set the desired target size in bytes.
                                        #Takes precedence over the 'compression' parameter.
        float target_PSNR               #if non-zero, specifies the minimal distortion to
                                        #try to achieve. Takes precedence over target_size.
        int segments                    #maximum number of segments to use, in [1..4]
        int sns_strength                #Spatial Noise Shaping. 0=off, 100=maximum.
        int filter_strength             #range: [0 = off .. 100 = strongest]
        int filter_sharpness            #range: [0 = off .. 7 = least sharp]
        int filter_type                 #filtering type: 0 = simple, 1 = strong (only used
                                        #if filter_strength > 0 or autofilter > 0)
        int autofilter                  #Auto adjust filter's strength [0 = off, 1 = on]
        int alpha_compression           #Algorithm for encoding the alpha plane (0 = none,
                                        #1 compressed with WebP lossless). Default is 1.
        int alpha_filtering             #Predictive filtering method for alpha plane.
                                        #0: none, 1: fast, 2: best. Default if 1.
        int alpha_quality               #Between 0 (smallest size) and 100 (lossless).
                                        #Default is 100.
        int _pass "pass"                #number of entropy-analysis passes (in [1..10]).

        int show_compressed             #if true, export the compressed picture back.
                                        #In-loop filtering is not applied.
        int preprocessing               #preprocessing filter (0=none, 1=segment-smooth)
        int partitions                  #log2(number of token partitions) in [0..3]. Default
                                        #is set to 0 for easier progressive decoding.
        int partition_limit             #quality degradation allowed to fit the 512k limit
                                        #on prediction modes coding (0: no degradation,
                                        #100: maximum possible degradation).
        int emulate_jpeg_size           #If true, compression parameters will be remapped
                                        #to better match the expected output size from
                                        #JPEG compression. Generally, the output size will
                                        #be similar but the degradation will be lower.
        int thread_level                #If non-zero, try and use multi-threaded encoding.
        int low_memory                  #If set, reduce memory usage (but increase CPU use).

        int near_lossless               #Near lossless encoding [0 = max loss .. 100 = off
                                        #(default)
        int exact                       #if non-zero, preserve the exact RGB values under
                                        #transparent area. Otherwise, discard this invisible
                                        #RGB information for better compression. The default
                                        #value is 0.
        uint32_t pad[3]                 #padding for later use

    ctypedef struct WebPMemoryWriter:
        uint8_t* mem                    #final buffer (of size 'max_size', larger than 'size').
        size_t   size                   #final size
        size_t   max_size               #total capacity
        uint32_t pad[1]

    ctypedef void *WebPWriterFunction
    ctypedef void *WebPProgressHook
    ctypedef void *WebPAuxStats

    ctypedef struct WebPPicture:
        #   INPUT
        # Main flag for encoder selecting between ARGB or YUV input.
        # It is recommended to use ARGB input (*argb, argb_stride) for lossless
        # compression, and YUV input (*y, *u, *v, etc.) for lossy compression
        # since these are the respective native colorspace for these formats.
        int use_argb

        # YUV input (mostly used for input to lossy compression)
        WebPEncCSP colorspace           #colorspace: should be YUV420 for now (=Y'CbCr).
        int width, height               #dimensions (less or equal to WEBP_MAX_DIMENSION)
        uint8_t *y
        uint8_t *u
        uint8_t *v
        int y_stride, uv_stride         #luma/chroma strides.
        uint8_t* a                      #pointer to the alpha plane
        int a_stride                    #stride of the alpha plane
        uint32_t pad1[2]                #padding for later use

        # ARGB input (mostly used for input to lossless compression)
        uint32_t* argb                  #Pointer to argb (32 bit) plane.
        int argb_stride                 #This is stride in pixels units, not bytes.
        uint32_t pad2[3]                #padding for later use

        #   OUTPUT
        # Byte-emission hook, to store compressed bytes as they are ready.
        WebPWriterFunction writer       #can be NULL
        void* custom_ptr                #can be used by the writer.

        # map for extra information (only for lossy compression mode)
        int extra_info_type             #1: intra type, 2: segment, 3: quant
                                        #4: intra-16 prediction mode,
                                        #5: chroma prediction mode,
                                        #6: bit cost, 7: distortion
        uint8_t* extra_info             #if not NULL, points to an array of size
                                        # ((width + 15) / 16) * ((height + 15) / 16) that
                                        #will be filled with a macroblock map, depending
                                        #on extra_info_type.

        #   STATS AND REPORTS
        # Pointer to side statistics (updated only if not NULL)
        WebPAuxStats* stats

        # Error code for the latest error encountered during encoding
        WebPEncodingError error_code

        #If not NULL, report progress during encoding.
        WebPProgressHook progress_hook

        void* user_data                 #this field is free to be set to any value and
                                        #used during callbacks (like progress-report e.g.).

        uint32_t pad3[3]                #padding for later use

        # Unused for now: original samples (for non-YUV420 modes)
        uint8_t *u0
        uint8_t *v0
        int uv0_stride

        uint32_t pad4[7]                #padding for later use

        # PRIVATE FIELDS
        void* memory_                   #row chunk of memory for yuva planes
        void* memory_argb_              #and for argb too.
        void* pad5[2]                   #padding for later use

    void WebPMemoryWriterInit(WebPMemoryWriter* writer)
    int WebPMemoryWrite(const uint8_t* data, size_t data_size, const WebPPicture* picture) nogil

    int WebPConfigInit(WebPConfig* config)
    int WebPConfigPreset(WebPConfig* config, WebPPreset preset, float quality)
    int WebPValidateConfig(const WebPConfig* config)
    int WebPPictureInit(WebPPicture* picture)
    void WebPPictureFree(WebPPicture* picture)

    # Colorspace conversion function to import RGB samples.
    # Previous buffer will be free'd, if any.
    # *rgb buffer should have a size of at least height * rgb_stride.
    # Returns false in case of memory error.
    int WebPPictureImportRGB(WebPPicture* picture, const uint8_t* rgb, int rgb_stride) nogil
    # Same, but for RGBA buffer.
    int WebPPictureImportRGBA(WebPPicture* picture, const uint8_t* rgba, int rgba_stride) nogil
    # Same, but for RGBA buffer. Imports the RGB direct from the 32-bit format
    # input buffer ignoring the alpha channel. Avoids needing to copy the data
    # to a temporary 24-bit RGB buffer to import the RGB only.
    int WebPPictureImportRGBX(WebPPicture* picture, const uint8_t* rgbx, int rgbx_stride) nogil

    # Variants of the above, but taking BGR(A|X) input.
    int WebPPictureImportBGR(WebPPicture* picture, const uint8_t* bgr, int bgr_stride) nogil
    int WebPPictureImportBGRA(WebPPicture* picture, const uint8_t* bgra, int bgra_stride) nogil
    int WebPPictureImportBGRX(WebPPicture* picture, const uint8_t* bgrx, int bgrx_stride) nogil

    # Converts picture->argb data to the YUVA format specified by 'colorspace'.
    # Upon return, picture->use_argb is set to false. The presence of real
    # non-opaque transparent values is detected, and 'colorspace' will be
    # adjusted accordingly. Note that this method is lossy.
    # Returns false in case of error.
    int WebPPictureARGBToYUVA(WebPPicture* picture, WebPEncCSP colorspace) nogil

    # Converts picture->yuv to picture->argb and sets picture->use_argb to true.
    # The input format must be YUV_420 or YUV_420A.
    # Note that the use of this method is discouraged if one has access to the
    # raw ARGB samples, since using YUV420 is comparatively lossy. Also, the
    # conversion from YUV420 to ARGB incurs a small loss too.
    # Returns false in case of error.
    int WebPPictureYUVAToARGB(WebPPicture* picture) nogil

    # Helper function: given a width x height plane of YUV(A) samples
    # (with stride 'stride'), clean-up the YUV samples under fully transparent
    # area, to help compressibility (no guarantee, though).
    void WebPCleanupTransparentArea(WebPPicture* picture) nogil

    # Scan the picture 'picture' for the presence of non fully opaque alpha values.
    # Returns true in such case. Otherwise returns false (indicating that the
    # alpha plane can be ignored altogether e.g.).
    int WebPPictureHasTransparency(const WebPPicture* picture) nogil

    # Main encoding call, after config and picture have been initialized.
    # 'picture' must be less than 16384x16384 in dimension (cf WEBP_MAX_DIMENSION),
    # and the 'config' object must be a valid one.
    # Returns false in case of error, true otherwise.
    # In case of error, picture->error_code is updated accordingly.
    # 'picture' can hold the source samples in both YUV(A) or ARGB input, depending
    # on the value of 'picture->use_argb'. It is highly recommended to use
    # the former for lossy encoding, and the latter for lossless encoding
    # (when config.lossless is true). Automatic conversion from one format to
    # another is provided but they both incur some loss.
    int WebPEncode(const WebPConfig* config, WebPPicture* picture) nogil

    #  Rescale a picture to new dimension width x height.
    # If either 'width' or 'height' (but not both) is 0 the corresponding
    # dimension will be calculated preserving the aspect ratio.
    # No gamma correction is applied.
    # Returns false in case of error (invalid parameter or insufficient memory).
    int WebPPictureRescale(WebPPicture* pic, int width, int height) nogil


ERROR_TO_NAME: Dict[int, str] = {
#VP8_ENC_OK
    VP8_ENC_ERROR_OUT_OF_MEMORY             : "memory error allocating objects",
    VP8_ENC_ERROR_BITSTREAM_OUT_OF_MEMORY   : "memory error while flushing bits",
    VP8_ENC_ERROR_NULL_PARAMETER            : "a pointer parameter is NULL",
    VP8_ENC_ERROR_INVALID_CONFIGURATION     : "configuration is invalid",
    VP8_ENC_ERROR_BAD_DIMENSION             : "picture has invalid width/height",
    VP8_ENC_ERROR_PARTITION0_OVERFLOW       : "partition is bigger than 512k",
    VP8_ENC_ERROR_PARTITION_OVERFLOW        : "partition is bigger than 16M",
    VP8_ENC_ERROR_BAD_WRITE                 : "error while flushing bytes",
    VP8_ENC_ERROR_FILE_TOO_BIG              : "file is bigger than 4G",
    VP8_ENC_ERROR_USER_ABORT                : "abort request by user",
}

PRESETS: Dict[int, str] = {
    WEBP_PRESET_DEFAULT      : "default",
    WEBP_PRESET_PICTURE      : "picture",
    WEBP_PRESET_PHOTO        : "photo",
    WEBP_PRESET_DRAWING      : "drawing",
    WEBP_PRESET_ICON         : "icon",
    WEBP_PRESET_TEXT         : "text",
}
PRESET_NAME_TO_CONSTANT: Dict[str, int] = {}
for k,v in PRESETS.items():
    PRESET_NAME_TO_CONSTANT[v] = k

CONTENT_TYPE_PRESET: Dict[str, int] = {
    "picture"   : WEBP_PRESET_PICTURE,
    "text"      : WEBP_PRESET_TEXT,
    "browser"   : WEBP_PRESET_TEXT,
}

CSP_NAMES: Dict[int, str] = {
    WEBP_YUV420         : "YUV420",
    WEBP_YUV420A        : "YUV420A",
    WEBP_CSP_UV_MASK    : "UV_MASK",
}

IMAGE_HINT: Dict[int, str] = {
    WEBP_HINT_DEFAULT     : "default",
    WEBP_HINT_PICTURE     : "picture",
    WEBP_HINT_PHOTO       : "photo",
    WEBP_HINT_GRAPH       : "graph",
}
HINT_NAME_TO_CONSTANT: Dict[str, int] = {}
for k,v in IMAGE_HINT.items():
    HINT_NAME_TO_CONSTANT[v] = k

CONTENT_TYPE_HINT: Dict[str, int] = {
    "picture"   : WEBP_HINT_PICTURE,
}

cdef WebPImageHint DEFAULT_IMAGE_HINT = HINT_NAME_TO_CONSTANT.get(os.environ.get("XPRA_WEBP_IMAGE_HINT", "default").lower(), WEBP_HINT_DEFAULT)
cdef WebPPreset DEFAULT_PRESET = PRESET_NAME_TO_CONSTANT.get(os.environ.get("XPRA_WEBP_PRESET", "default").lower(), WEBP_PRESET_DEFAULT)
cdef WebPPreset PRESET_SMALL = PRESET_NAME_TO_CONSTANT.get(os.environ.get("XPRA_WEBP_PRESET_SMALL", "icon").lower(), WEBP_PRESET_ICON)


def get_type() -> str:
    return "webp"


def get_encodings() -> Sequence[str]:
    return ("webp", )


def get_version() -> Tuple[int, int, int]:
    cdef int version = WebPGetEncoderVersion()
    log("WebPGetEncoderVersion()=%#x", version)
    return (version >> 16) & 0xff, (version >> 8) & 0xff, version & 0xff


def get_info() -> Dict[str, Any]:
    return  {
        "version"       : get_version(),
        "encodings"     : get_encodings(),
        "threading"     : bool(WEBP_THREADING),
        "image-hint"    : DEFAULT_IMAGE_HINT,
        "image-hints"   : tuple(IMAGE_HINT.values()),
        "preset"        : DEFAULT_PRESET,
        "preset-small"  : PRESET_SMALL,
        "presets"       : tuple(PRESETS.values()),
    }


INPUT_PIXEL_FORMATS = ("RGBX", "RGBA", "BGRX", "BGRA", "RGB", "BGR")


def get_specs() -> Sequence[VideoSpec]:
    specs: Sequence[VideoSpec] = []
    for in_cs in INPUT_PIXEL_FORMATS:
        out_cs = in_cs
        specs.append(VideoSpec(
                encoding="webp", input_colorspace=in_cs, output_colorspaces=(out_cs, ),
                has_lossless_mode=False,
                codec_class=Encoder, codec_type="webp",
                setup_cost=0, cpu_cost=100, gpu_cost=0,
                min_w=16, min_h=16, max_w=4*1024, max_h=4*1024,
                can_scale=True,
                score_boost=-50,
            )
        )
    return specs


cdef inline webp_check(int ret):
    if ret==0:
        return
    err = ERROR_TO_NAME.get(ret, ret)
    raise RuntimeError("error: %s" % err)


cdef inline float fclamp(int v):
    if v<0:
        v = 0
    elif v>100:
        v = 100
    return <float> v


cdef dict get_config_info(WebPConfig *config):
    return {
        "lossless"          : config.lossless,
        "method"            : config.method,
        "image_hint"        : IMAGE_HINT.get(config.image_hint, config.image_hint),
        "target_size"       : config.target_size,
        "target_PSNR"       : config.target_PSNR,
        "segments"          : config.segments,
        "sns_strength"      : config.sns_strength,
        "filter_strength"   : config.filter_strength,
        "filter_sharpness"  : config.filter_sharpness,
        "filter_type"       : config.filter_type,
        "autofilter"        : config.autofilter,
        "alpha_compression" : config.alpha_compression,
        "alpha_filtering"   : config.alpha_filtering,
        "alpha_quality"     : config.alpha_quality,
        "pass"              : config._pass,
        "show_compressed"   : config.show_compressed,
        "preprocessing"     : config.preprocessing,
        "partitions"        : config.partitions,
        "partition_limit"   : config.partition_limit,
        "emulate_jpeg_size" : config.emulate_jpeg_size,
        "thread_level"      : config.thread_level,
        "low_memory"        : config.low_memory,
    }



cdef class Encoder:
    cdef int width
    cdef int height
    cdef object src_format
    cdef unsigned int Bpp
    cdef int quality
    cdef int speed
    cdef unsigned char alpha
    cdef object content_type
    cdef long frames
    cdef WebPConfig config
    cdef WebPPreset preset
    cdef object __weakref__

    def __init__(self):
        self.width = self.height = self.quality = self.frames = 0

    def init_context(self, encoding: str, width : int, height : int, src_format: str, options: typedict) -> None:
        assert encoding=="webp", "invalid encoding: %s" % encoding
        assert src_format in INPUT_PIXEL_FORMATS
        self.width = width
        self.height = height
        self.src_format = src_format
        self.Bpp = len(src_format)     #ie: "BGRA" -> 4
        self.quality = options.intget("quality", 50)
        self.speed = options.intget("speed", 50)
        self.alpha = src_format.find("A")>=0
        self.content_type = options.get("content-type", "")
        self.configure_encoder()

    cdef void configure_encoder(self):
        cdef int ret = WebPConfigInit(&self.config)
        if not ret:
            raise RuntimeError("failed to initialize webp config")
        config_init(&self.config)
        self.preset = get_preset(self.width, self.height, self.content_type)
        configure_preset(&self.config, self.preset, self.quality)
        configure_encoder(&self.config, self.quality, self.speed, self.alpha)
        configure_image_hint(&self.config, self.content_type)
        validate_config(&self.config)

    def is_ready(self) -> bool:
        return self.width>0 and self.height>0

    def is_closed(self) -> bool:
        return self.width==0 or self.height==0

    def clean(self) -> None:
        self.width = self.height = self.quality = 0

    def get_encoding(self) -> str:
        return "webp"

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_type(self) -> str:
        return "webp"

    def get_src_format(self) -> str:
        return self.src_format

    def get_info(self) -> Dict[str,Any]:
        info = get_info()
        info |= {
            "frames"        : int(self.frames),
            "width"         : self.width,
            "height"        : self.height,
            "alpha"         : bool(self.alpha),
            "pixel-format"  : self.src_format,
            "content-type"  : self.content_type,
        }
        return info

    def compress_image(self, image: ImageWrapper, options: typedict) -> Tuple:
        reconfigure = False
        quality = options.get("quality", -1)
        speed = options.get("speed", -1)
        pixel_format = image.get_pixel_format()
        if quality>0 and quality!=self.quality:
            self.quality = quality
            reconfigure = True
        if speed>0 and speed!=self.speed:
            self.speed = speed
            reconfigure = True
        if image.get_width()!=self.width or image.get_height()!=self.height:
            self.width = image.get_width()
            self.height = image.get_height()
            reconfigure = True
        if pixel_format!=self.src_format:
            self.src_format = pixel_format
            self.Bpp = len(pixel_format)
            self.alpha = pixel_format.find("A")>=0
            reconfigure = True
        if options.get("content-type", "") != self.content_type:
            self.content_type = options.get("content-type", "")
            reconfigure = True
        if reconfigure:
            log("webp reconfigure")
            self.configure_encoder()

        cdef unsigned int stride = image.get_rowstride()
        pixels = image.get_pixels()
        cdef WebPPicture pic
        import_picture(&pic, self.width, self.height,
                       stride,
                       self.alpha,
                       pixel_format, pixels)

        cdef int scaled_width = options.intget("scaled-width", self.width)
        cdef int scaled_height = options.intget("scaled-height", self.height)
        if scaled_width!=self.width or scaled_height!=self.height:
            scale_picture(&pic, self.scaled_width, self.scaled_height)

        client_options = {
            "rgb_format"  : pixel_format,
            "frame" : self.frames,
        }
        self.frames += 1
        cdef int subsample = self.quality < SUBSAMPLING_THRESHOLD
        if subsample:
            if self.alpha:
                to_yuv(&pic, WEBP_YUV420A)
            else:
                to_yuv(&pic, WEBP_YUV420)
            client_options["subsampling"] = "YUV420P"

        cdata = webp_encode(&self.config, &pic)

        if self.config.lossless:
            client_options["quality"] = 100
        else:
            client_options["quality"] = max(0, min(99, quality))
        if self.alpha:
            client_options["has_alpha"] = True
        log("webp compression ratio=%2i%%, client-options=%s",
            100*len(cdata)//(self.width*self.height*self.Bpp), client_options)
        if LOG_CONFIG>0:
            log("webp.compress used config: %s", get_config_info(&self.config))
        may_save_image("webp", cdata)
        return cdata, client_options


cdef inline WebPPreset get_preset(unsigned int width, unsigned int height, content_type: str):
    cdef WebPPreset preset = DEFAULT_PRESET
    #only use icon for small squarish rectangles
    if width*height<=2304 and abs(width-height)<=16:
        preset = PRESET_SMALL
    return CONTENT_TYPE_PRESET.get(content_type, preset)


cdef inline void config_init(WebPConfig *config):
    cdef int ret = WebPConfigInit(config)
    if not ret:
        raise RuntimeError("failed to initialize webp config")


cdef configure_encoder(WebPConfig *config,
                      unsigned int quality, unsigned int speed,
                      unsigned char alpha):
    config.lossless = quality>=100
    if config.lossless:
        #'quality' actually controls the speed
        #and anything above zero is just too slow:
        config.quality = 0
        config.autofilter = 1
    else:
        config.quality = fclamp(quality)
        config.segments = 1
        config.sns_strength = 0
        config.filter_strength = 0
        config.filter_sharpness = 7-quality//15
        config.filter_type = 0
        config.autofilter = 0
    #"method" takes values from 0 to 6,
    #but anything higher than 1 is dreadfully slow,
    #so only use method=1 when speed is already very low
    config.method = int(speed<10)
    config.alpha_compression = alpha
    config.alpha_filtering = MAX(0, MIN(2, speed/50)) * alpha
    config.alpha_quality = quality * alpha
    config.emulate_jpeg_size = 1
    config._pass = MAX(1, MIN(10, (40-speed)//10))
    config.preprocessing = int(speed<30)
    config.thread_level = WEBP_THREADING
    config.partitions = 3
    config.partition_limit = MAX(0, MIN(100, 100-quality))
    log("webp.configure_encoder config: lossless=%-5s, quality=%3i, method=%i, alpha=%3i,%3i,%3i",
        config.lossless, config.quality, config.method,
        config.alpha_compression, config.alpha_filtering, config.alpha_quality)


cdef void configure_preset(WebPConfig *config, WebPPreset preset, int quality):
    ret = WebPConfigPreset(config, preset, fclamp(quality))
    if not ret:
        raise ValueError(f"failed to set webp preset {preset} and quality {quality}")
    log("webp config: preset=%-8s", PRESETS.get(preset, preset))


cdef void configure_image_hint(WebPConfig *config, content_type: str):
    cdef WebPImageHint image_hint = CONTENT_TYPE_HINT.get(content_type, DEFAULT_IMAGE_HINT)
    config.image_hint = image_hint
    log("webp config: image hint=%s", IMAGE_HINT.get(image_hint, image_hint))


cdef void validate_config(WebPConfig *config):
    ret = WebPValidateConfig(config)
    if not ret:
        info = get_config_info(config)
        raise RuntimeError("invalid webp configuration: %s" % info)


def encode(coding: str, image: ImageWrapper, options: typedict) -> Tuple:
    log("webp.encode(%s, %s, %s)", coding, image, options)
    assert coding=="webp"
    pixel_format = image.get_pixel_format()
    if pixel_format not in INPUT_PIXEL_FORMATS:
        raise ValueError(f"unsupported pixel format {pixel_format!r}")
    options = options or {}

    cdef unsigned int width = image.get_width()
    cdef unsigned int height = image.get_height()
    cdef unsigned int stride = image.get_rowstride()
    assert width<16384 and height<16384, "invalid image dimensions: %ix%i" % (width, height)
    cdef unsigned int scaled_width = options.intget("scaled-width", width)
    cdef unsigned int scaled_height = options.intget("scaled-height", height)
    assert scaled_width<16384 and scaled_height<16384, "invalid image dimensions: %ix%i" % (width, height)
    cdef unsigned int Bpp = len(pixel_format)   #ie: "BGRA" -> 4
    cdef unsigned int supports_alpha = options.boolget("alpha", True)
    cdef unsigned char alpha = supports_alpha and pixel_format.find("A")>=0
    cdef int quality = options.intget("quality", 50)
    cdef int speed = options.intget("speed", 50)

    cdef WebPConfig config
    config_init(&config)

    content_type = options.strget("content-type", "")
    cdef WebPPreset preset = get_preset(width, height, content_type)
    configure_preset(&config, preset, quality)
    configure_encoder(&config, quality, speed, alpha)
    configure_image_hint(&config, content_type)
    validate_config(&config)

    pixels = image.get_pixels()
    cdef WebPPicture pic
    import_picture(&pic, width, height,
                   stride,
                   supports_alpha,
                   pixel_format, pixels)

    if scaled_width!=width or scaled_height!=height:
        scale_picture(&pic, scaled_width, scaled_height)

    client_options = {
        "rgb_format"  : pixel_format,
    }
    cdef int subsample = quality < SUBSAMPLING_THRESHOLD
    if subsample:
        if alpha:
            to_yuv(&pic, WEBP_YUV420A)
        else:
            to_yuv(&pic, WEBP_YUV420)
        client_options["subsampling"] = "YUV420P"

    cdata = webp_encode(&config, &pic)

    if config.lossless:
        client_options["quality"] = 100
    else:
        client_options["quality"] = max(0, min(99, quality))
    if alpha:
        client_options["has_alpha"] = True
    log("webp.compress ratio=%i%%, client-options=%s", 100*len(cdata)//(width*height*Bpp), client_options)
    if LOG_CONFIG>0:
        log("webp.compress used config: %s", get_config_info(&config))
    may_save_image("webp", cdata)
    return "webp", Compressed("webp", cdata), client_options, width, height, 0, len(pixel_format.replace("A", ""))*8


cdef void import_picture(WebPPicture *pic,
                  unsigned int width, unsigned int height,
                  unsigned int stride,
                  unsigned char supports_alpha,
                  pixel_format: str, pixels):
    memset(pic, 0, sizeof(WebPPicture))
    ret = WebPPictureInit(pic)
    if not ret:
        raise RuntimeError("failed to initialise webp picture")
    cdef unsigned int Bpp = len(pixel_format)   #ie: "BGRA" -> 4
    pic.width = width
    pic.height = height
    pic.use_argb = 1
    pic.argb_stride = stride//Bpp
    cdef int size = stride * height
    cdef double start = monotonic()
    cdef double end
    cdef const uint8_t* src
    with buffer_context(pixels) as bc:
        assert len(bc)>=size, "pixel buffer is too small: expected at least %s bytes but got %s" % (size, len(bc))
        src = <const uint8_t*> (<uintptr_t> int(bc))

        #import the pixel data into WebPPicture
        if pixel_format=="RGB":
            with nogil:
                ret = WebPPictureImportRGB(pic, src, stride)
        elif pixel_format=="BGR":
            with nogil:
                ret = WebPPictureImportBGR(pic, src, stride)
        elif pixel_format=="RGBX" or (pixel_format=="RGBA" and not supports_alpha):
            with nogil:
                ret = WebPPictureImportRGBX(pic, src, stride)
        elif pixel_format=="RGBA":
            with nogil:
                ret = WebPPictureImportRGBA(pic, src, stride)
        elif pixel_format=="BGRX" or (pixel_format=="BGRA" and not supports_alpha):
            with nogil:
                ret = WebPPictureImportBGRX(pic, src, stride)
        else:
            assert pixel_format=="BGRA"
            with nogil:
                ret = WebPPictureImportBGRA(pic, src, stride)
    if not ret:
        WebPPictureFree(pic)
        raise RuntimeError("WebP importing image failed: %s" % (ERROR_TO_NAME.get(pic.error_code, pic.error_code)))
    end = monotonic()
    log("webp %s import took %.1fms", pixel_format, 1000*(end-start))


cdef void scale_picture(WebPPicture *pic, unsigned scaled_width, unsigned int scaled_height):
    cdef double start = monotonic()
    with nogil:
        ret = WebPPictureRescale(pic, scaled_width, scaled_height)
    if not ret:
        WebPPictureFree(pic)
        raise RuntimeError("WebP failed to resize to %ix%i" % (scaled_width, scaled_height))
    cdef double end = monotonic()
    log("webp %s resizing took %.1fms", 1000*(end-start))


cdef void to_yuv(WebPPicture *pic, WebPEncCSP csp=WEBP_YUV420):
    cdef double start = monotonic()
    with nogil:
        ret = WebPPictureARGBToYUVA(pic, csp)
    if not ret:
        raise RuntimeError("WebPPictureARGBToYUVA failed")
    cdef double end = monotonic()
    log("webp subsampling ARGB to %s took %.1fms", CSP_NAMES.get(csp, csp), 1000*(end-start))


cdef object webp_encode(WebPConfig *config, WebPPicture *pic):
    cdef double start = monotonic()
    cdef WebPMemoryWriter memory_writer
    memset(&memory_writer, 0, sizeof(WebPMemoryWriter))
    try:
        #TODO: custom writer that over-allocates memory
        WebPMemoryWriterInit(&memory_writer)
        pic.writer = <WebPWriterFunction> WebPMemoryWrite
        pic.custom_ptr = <void*> &memory_writer
        with nogil:
            ret = WebPEncode(config, pic)
        if not ret:
            raise RuntimeError("WebPEncode failed: %s, config=%s" % (
                ERROR_TO_NAME.get(pic.error_code, pic.error_code), get_config_info(config)))
        cdata = memory_writer.mem[:memory_writer.size]
    finally:
        if memory_writer.mem:
            free(memory_writer.mem)
        WebPPictureFree(pic)
    cdef double end = monotonic()
    log("webp encode took %.1fms", 1000*(end-start))
    return cdata


def selftest(full=False) -> None:
    #fake empty buffer:
    from xpra.codecs.checks import make_test_image
    w, h = (24, 16)
    for has_alpha in (True, False):
        img = make_test_image("BGR%s" % ["X", "A"][has_alpha], w, h)
        for q in (10, 50, 90):
            r = encode("webp", img, typedict({"quality" : q, "speed" : 50, "alpha" : has_alpha}))
            assert len(r)>0
        #import binascii
        #print("compressed data(%s)=%s" % (has_alpha, binascii.hexlify(r)))



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False

import os
from time import monotonic
from typing import Any, Tuple, Dict
from collections.abc import Sequence

from xpra.log import Logger
log = Logger("encoder", "x264")

from xpra.util.env import envint, envbool
from xpra.util.str_fn import csv
from xpra.util.objects import typedict, AtomicInteger
from xpra.codecs.image import ImageWrapper
from xpra.codecs.constants import VideoSpec, get_profile, get_x264_quality, get_x264_preset
from collections import deque

from libc.string cimport memset
from libc.stdint cimport int64_t, uint64_t, uint8_t, uintptr_t
from xpra.buffers.membuf cimport memalign


MB_INFO = envbool("XPRA_X264_MB_INFO", False)
MAX_DELAYED_FRAMES = envint("XPRA_X264_MAX_DELAYED_FRAMES", 4)
THREADS = envint("XPRA_X264_THREADS", min(4, max(1, os.cpu_count()//2)))
MIN_SLICED_THREADS_SPEED = envint("XPRA_X264_SLICED_THREADS", 60)
LOGGING = os.environ.get("XPRA_X264_LOGGING", "WARNING")
SUPPORT_24BPP = envbool("XPRA_X264_SUPPORT_24BPP")
SUPPORT_30BPP = envbool("XPRA_X264_SUPPORT_30BPP", True)
TUNE = os.environ.get("XPRA_X264_TUNE")
LOG_NALS = envbool("XPRA_X264_LOG_NALS")
SAVE_TO_FILE = os.environ.get("XPRA_SAVE_TO_FILE")
BLANK_VIDEO = envbool("XPRA_X264_BLANK_VIDEO")


DEF X264_MBINFO_CONSTANT = 1


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "stdlib.h":
    void free(void *ptr)

cdef extern from "string.h":
    int vsnprintf(char * s, size_t n, const char *fmt, va_list arg)

cdef extern from "stdarg.h":
    ctypedef struct va_list:
        pass
    ctypedef struct fake_type:
        pass
    void va_start(va_list, void* arg)
    void* va_arg(va_list, fake_type)
    void va_end(va_list)
    fake_type int_type "int"

cdef extern from "x264.h":
    int X264_KEYINT_MAX_INFINITE

    int X264_BUILD

    int X264_LOG_DEBUG
    int X264_LOG_INFO
    int X264_LOG_WARNING
    int X264_LOG_ERROR

    int X264_CSP_I420
    int X264_CSP_I422
    int X264_CSP_I444
    int X264_CSP_BGR
    int X264_CSP_BGRA
    int X264_CSP_RGB
    int X264_CSP_NV12
    int X264_CSP_V210
    int X264_CSP_HIGH_DEPTH

    int X264_RC_CQP
    int X264_RC_CRF
    int X264_RC_ABR

    int X264_B_ADAPT_NONE
    int X264_B_ADAPT_FAST
    int X264_B_ADAPT_TRELLIS

    #enum
    int X264_ME_DIA
    int X264_ME_HEX
    int X264_ME_UMH
    int X264_ME_ESA
    int X264_ME_TESA

    #enum nal_unit_type_e
    int NAL_UNKNOWN
    int NAL_SLICE
    int NAL_SLICE_DPA
    int NAL_SLICE_DPB
    int NAL_SLICE_DPC
    int NAL_SLICE_IDR
    int NAL_SEI
    int NAL_SPS
    int NAL_PPS
    int NAL_AUD
    int NAL_FILLER

    #enum nal_priority_e
    int NAL_PRIORITY_DISPOSABLE
    int NAL_PRIORITY_LOW
    int NAL_PRIORITY_HIGH
    int NAL_PRIORITY_HIGHEST

    #frame type
    int X264_TYPE_AUTO              # Let x264 choose the right type
    int X264_TYPE_KEYFRAME
    int X264_TYPE_IDR
    int X264_TYPE_I
    int X264_TYPE_P
    int X264_TYPE_BREF
    int X264_TYPE_B

    int X264_WEIGHTP_NONE

    const char * const *x264_preset_names

    ctypedef struct rc:
        int         i_rc_method
        int         i_qp_constant       #0 to (51 + 6*(x264_bit_depth-8)). 0=lossless
        int         i_qp_min            #min allowed QP value
        int         i_qp_max            #max allowed QP value
        int         i_qp_step           #max QP step between frames

        int         i_bitrate
        float       f_rf_constant       #1pass VBR, nominal QP
        float       f_rf_constant_max   #In CRF mode, maximum CRF as caused by VBV
        float       f_rate_tolerance
        int         i_vbv_max_bitrate
        int         i_vbv_buffer_size
        float       f_vbv_buffer_init   #<=1: fraction of buffer_size. >1: kbit
        float       f_ip_factor
        float       f_pb_factor

        int         i_aq_mode           #psy adaptive QP. (X264_AQ_*)
        float       f_aq_strength
        int         b_mb_tree           #Macroblock-tree ratecontrol
        int         i_lookahead

        # 2pass
        int         b_stat_write        #Enable stat writing in psz_stat_out
        char        *psz_stat_out       #output filename (in UTF-8) of the 2pass stats file
        int         b_stat_read         #Read stat from psz_stat_in and use it
        char        *psz_stat_in        #input filename (in UTF-8) of the 2pass stats file

        # 2pass params (same as ffmpeg ones)
        float       f_qcompress         #0.0 => cbr, 1.0 => constant qp
        float       f_qblur             #temporally blur quants
        float       f_complexity_blur   #temporally blur complexity
        #x264_zone_t *zones              #ratecontrol overrides
        int         i_zones             #number of zone_t's
        char        *psz_zones          #alternate method of specifying zones

    ctypedef struct analyse:
        int         intra
        int         inter
        int         b_fast_pskip
        int         b_mb_info
        int         i_me_method         # motion estimation algorithm to use (X264_ME_*)
        int         i_me_range          # integer pixel motion estimation search range (from predicted mv) */
        int         i_mv_range          # maximum length of a mv (in pixels). -1 = auto, based on level */
        int         i_mv_range_thread   # minimum space between threads. -1 = auto, based on number of threads. */
        int         i_subpel_refine     # subpixel motion estimation quality */
        int         i_weighted_pred     # weighting for P-frames
        int         b_weighted_bipred   # implicit weighting for B-frames

    ctypedef struct vui:
        int         i_sar_height
        int         i_sar_width
        int         i_overscan          # 0=undef, 1=no overscan, 2=overscan
        # see h264 annex E for the values of the following
        int         i_vidformat
        int         b_fullrange
        int         i_colorprim
        int         i_transfer
        int         i_colmatrix
        int         i_chroma_loc        # both top & bottom

    ctypedef struct x264_param_t:
        unsigned int cpu
        int i_threads           #encode multiple frames in parallel
        int i_lookahead_threads #multiple threads for lookahead analysis
        int b_sliced_threads    #Whether to use slice-based threading
        int b_deterministic     #whether to allow non-deterministic optimizations when threaded
        int b_cpu_independent   #force canonical behavior rather than cpu-dependent optimal algorithms
        int i_sync_lookahead    #threaded lookahead buffer

        int i_width
        int i_height
        int i_csp               #CSP of encoded bitstream
        int i_bitdepth
        int i_level_idc
        int i_frame_total       #number of frames to encode if known, else 0

        vui vui

        int i_log_level
        void* pf_log

        #Bitstream parameters
        int i_frame_reference   #Maximum number of reference frames
        int i_dpb_size          #Force a DPB size larger than that implied by B-frames and reference frames
                                #Useful in combination with interactive error resilience.
        int i_keyint_max        #Force an IDR keyframe at this interval
        int i_keyint_min        #Scenecuts closer together than this are coded as I, not IDR.
        int i_scenecut_threshold#how aggressively to insert extra I frames
        int b_intra_refresh     #Whether or not to use periodic intra refresh instead of IDR frames.

        int i_bframe            #how many b-frame between 2 references pictures
        int i_bframe_adaptive
        int i_bframe_bias
        int i_bframe_pyramid    #Keep some B-frames as references: 0=off, 1=strict hierarchical, 2=normal
        int b_open_gop
        int b_bluray_compat
        #older x264 builds do not support this:
        #int b_opencl            #use OpenCL when available

        int b_cabac
        int b_deblocking_filter
        int b_interlaced
        int b_constrained_intra

        rc  rc                  #rate control
        analyse analyse

        int b_vfr_input         #VFR input. If 1, use timebase and timestamps for ratecontrol purposes. If 0, use fps only

    ctypedef struct x264_t:
        pass
    ctypedef struct x264_nal_t:
        int i_ref_idc
        int i_type
        int b_long_startcode
        int i_first_mb
        int i_last_mb
        int i_payload
        uint8_t *p_payload
    ctypedef struct x264_image_t:
        int i_csp           #Colorspace
        int i_plane         #Number of image planes
        int i_stride[4]     #Strides for each plane
        uint8_t *plane[4]   #Pointers to each plane
    ctypedef struct x264_image_properties_t:
        float *quant_offsets
        void (*quant_offsets_free)(void*)
        uint8_t *mb_info
        void (*mb_info_free)(void*)
        double f_ssim
        double f_psnr_avg
        double f_psnr[3]
        double f_crf_avg
    ctypedef struct x264_hrd_t:
        pass
    ctypedef struct x264_sei_t:
        pass
    ctypedef struct x264_picture_t:
        int i_type          #In: force picture type (if not auto)
        int i_qpplus1       #In: force quantizer for != X264_QP_AUTO
        int i_pic_struct    #In: pic_struct, for pulldown/doubling/etc...used only if b_pic_struct=1.
                            #use pic_struct_e for pic_struct inputs
                            #Out: pic_struct element associated with frame
        int b_keyframe      #Out: whether this frame is a keyframe.  Important when using modes that result in
                            #SEI recovery points being used instead of IDR frames.
        int64_t i_pts       #In: user pts, Out: pts of encoded picture (user)
                            #Out: frame dts. When the pts of the first frame is close to zero,
                            #initial frames may have a negative dts which must be dealt with by any muxer
        x264_param_t *param #In: custom encoding parameters to be set from this frame forwards (..)
        x264_image_t img    #In: raw image data
                            #Out: Out: reconstructed image data
        x264_image_properties_t prop    #In: optional information to modify encoder decisions for this frame
                            #Out: information about the encoded frame */
        x264_hrd_t hrd_timing   #Out: HRD timing information. Output only when i_nal_hrd is set.
        x264_sei_t extra_sei#In: arbitrary user SEI (e.g subtitles, AFDs)
        void *opaque        #private user data. copied from input to output frames.

    void x264_picture_init(x264_picture_t *pic) nogil

    int x264_param_default_preset(x264_param_t *param, const char *preset, const char *tune)
    int x264_param_apply_profile(x264_param_t *param, const char *profile)
    void x264_encoder_parameters(x264_t *context, x264_param_t *param)
    int x264_encoder_reconfig(x264_t *context, x264_param_t *param)

    x264_t *x264_encoder_open(x264_param_t *param)
    void x264_encoder_close(x264_t *context)

    int x264_encoder_encode(x264_t *context, x264_nal_t **pp_nal, int *pi_nal, x264_picture_t *pic_in, x264_picture_t *pic_out ) nogil
    int x264_encoder_delayed_frames(x264_t *)
    int x264_encoder_maximum_delayed_frames(x264_t *h)


cdef set_f_rf(x264_param_t *param, float q):
    param.rc.f_rf_constant = q


cdef const char * const *get_preset_names():
    return x264_preset_names;


cdef inline int roundup(int n, int m):
    return (n + m - 1) & ~(m - 1)


ADAPT_TYPES: Dict[int, str] = {
    X264_B_ADAPT_NONE        : "NONE",
    X264_B_ADAPT_FAST        : "FAST",
    X264_B_ADAPT_TRELLIS     : "TRELLIS",
}

RC_TYPES: Dict[int, str] = {
    X264_RC_CQP : "CQP",
    X264_RC_CRF : "CRF",
    X264_RC_ABR : "ABR",
}

SLICE_TYPES: Dict[int, str] = {
    X264_TYPE_AUTO  : "auto",
    X264_TYPE_IDR   : "IDR",
    X264_TYPE_I     : "I",
    X264_TYPE_P     : "P",
    X264_TYPE_BREF  : "BREF",
    X264_TYPE_B     : "B",
    X264_TYPE_KEYFRAME  : "KEYFRAME",
}

ME_TYPES: Dict[int, str] = {
    X264_ME_DIA     : "DIA",
    X264_ME_HEX     : "HEX",
    X264_ME_UMH     : "UMH",
    X264_ME_ESA     : "ESA",
    X264_ME_TESA    : "TESA",
}

NAL_TYPES: Dict[int, str] = {
    NAL_UNKNOWN     : "unknown",
    NAL_SLICE       : "slice",
    NAL_SLICE_DPA   : "slice-dpa",
    NAL_SLICE_DPB   : "slice-dpb",
    NAL_SLICE_DPC   : "slice-dpc",
    NAL_SLICE_IDR   : "slice-idr",
    NAL_SEI         : "sei",
    NAL_SPS         : "sps",
    NAL_PPS         : "pps",
    NAL_AUD         : "aud",
    NAL_FILLER      : "filler",
}

NAL_PRIORITIES: Dict[int, str] = {
    NAL_PRIORITY_DISPOSABLE : "disposable",
    NAL_PRIORITY_LOW        : "low",
    NAL_PRIORITY_HIGH       : "high",
    NAL_PRIORITY_HIGHEST    : "highest",
}


PROFILE_BASELINE = "baseline"
PROFILE_MAIN     = "main"
PROFILE_HIGH     = "high"
PROFILE_HIGH10   = "high10"
PROFILE_HIGH422  = "high422"
PROFILE_HIGH444  = "high444"
I420_PROFILES = [PROFILE_BASELINE, PROFILE_MAIN, PROFILE_HIGH, PROFILE_HIGH10, PROFILE_HIGH422, PROFILE_HIGH444]
I422_PROFILES = [PROFILE_HIGH422, PROFILE_HIGH444]
I444_PROFILES = [PROFILE_HIGH444]
RGB_PROFILES = [PROFILE_HIGH444]

COLORSPACE_FORMATS : Dict[str,tuple] = {
    "YUV420P"   : (X264_CSP_I420,    PROFILE_HIGH,          I420_PROFILES),
    "YUV422P"   : (X264_CSP_I422,    PROFILE_HIGH422,       I422_PROFILES),
    "YUV444P"   : (X264_CSP_I444,    PROFILE_HIGH444,       I444_PROFILES),
    "BGRX"      : (X264_CSP_BGRA,    PROFILE_HIGH444,       RGB_PROFILES),
}
if SUPPORT_24BPP:
    COLORSPACE_FORMATS |= {
        "BGR"       : (X264_CSP_BGR,     PROFILE_HIGH444,    RGB_PROFILES),
        "RGB"       : (X264_CSP_RGB,     PROFILE_HIGH444,    RGB_PROFILES),
    }

COLORSPACES : Dict[str,str] = {
    "YUV420P"   : "YUV420P",
    "YUV422P"   : "YUV422P",
    "YUV444P"   : "YUV444P",
    "BGRX"      : "BGRX",
}
if SUPPORT_30BPP:
    COLORSPACE_FORMATS["BGR48"] = (X264_CSP_BGR | X264_CSP_HIGH_DEPTH,    PROFILE_HIGH444,    RGB_PROFILES)
    COLORSPACES["BGR48"] = "GBRP10"
if SUPPORT_24BPP:
    COLORSPACES |= {
        "BGR"       : "BGR",
        "RGB"       : "RGB",
    }


cdef inline bytes b(value: str):
    return value.encode("latin1")


def get_version() -> Tuple[int]:
    return (X264_BUILD, )


def get_type() -> str:
    return "x264"


generation = AtomicInteger()


def get_info() -> Dict[str,Any]:
    global COLORSPACES, MAX_WIDTH, MAX_HEIGHT
    return {
        "version"   : get_version(),
        "max-size"  : (MAX_WIDTH, MAX_HEIGHT),
        "generation": generation.get(),
        "formats"   : tuple(COLORSPACES.keys()),
        }


def get_encodings() -> Sequence[str]:
    return ("h264", )


#actual limits (which we cannot reach because we hit OOM):
#MAX_WIDTH, MAX_HEIGHT = (16384, 16384)
MAX_WIDTH, MAX_HEIGHT = (8192, 4096)


def get_specs() -> Sequence[VideoSpec]:
    specs: Sequence[VideoSpec] = []

    for in_cs, out_cs in COLORSPACES.items():
        # can handle high quality and any speed
        # setup cost is moderate (about 10ms)
        has_lossless_mode = in_cs in ("YUV444P", "BGR", "BGRX", "RGB")
        height_mask = width_mask = 0xFFFF
        if in_cs not in ("YUV444P", "BGRX"):
            width_mask = 0xFFFE
            if in_cs == "YUV420P":
                height_mask = 0xFFFE
        specs.append(VideoSpec(
                encoding="h264", input_colorspace=in_cs, output_colorspaces=(out_cs, ),
                has_lossless_mode=has_lossless_mode,
                codec_class=Encoder, codec_type=get_type(),
                quality=60+40*int(has_lossless_mode), speed=60,
                size_efficiency=60,
                setup_cost=20, width_mask=width_mask, height_mask=height_mask,
                max_w=MAX_WIDTH, max_h=MAX_HEIGHT,
            )
        )
    return specs


# maps a log level to one of our logger functions:
LOGGERS: Dict[int,Callable] = {
       X264_LOG_ERROR   : log.error,
       X264_LOG_WARNING : log.warn,
       X264_LOG_INFO    : log.info,
       X264_LOG_DEBUG   : log.debug,
}

# maps a log level string to the actual constant:
X264_LOG_MAP: Dict[str, int] = {
    "ERROR"    : X264_LOG_ERROR,
    "WARNING"  : X264_LOG_WARNING,
    "WARN"     : X264_LOG_WARNING,
    "INFO"     : X264_LOG_INFO,
    # getting segfaults with "DEBUG" level logging...
    # so this is currently disabled
    # "DEBUG"    : X264_LOG_DEBUG,
}
LOG_LEVEL: int = X264_LOG_MAP.get(LOGGING.upper(), X264_LOG_WARNING)


cdef str s(const char *v):
    pytmp = v[:]
    try:
        return pytmp.decode()
    except:
        return str(v[:])


# the static logging function we want x264 to use:
cdef void X264_log(void *p_unused, int level, const char *psz_fmt, va_list arg) noexcept with gil:
    cdef char buf[256]
    cdef int r = vsnprintf(buf, 256, psz_fmt, arg)
    if r < 0:
        log.error("X264_log: vsnprintf returned %s on format string '%s'", r, psz_fmt)
        return
    pystr = s(buf[:r]).rstrip("\n\r")
    logger = LOGGERS.get(level, log.info)
    logger("X264: %r", pystr)


cdef class Encoder:
    cdef unsigned long frames
    cdef x264_t *context
    cdef unsigned int width
    cdef unsigned int height
    cdef unsigned int fast_decode
    #cdef int opencl
    cdef object src_format
    cdef object csc_format
    cdef object content_type
    cdef object profile
    cdef object tune
    cdef double time
    cdef int colorspace
    cdef int preset
    cdef int quality
    cdef int speed
    cdef int b_frames
    cdef int max_delayed
    cdef int delayed_frames
    cdef int export_nals
    cdef int full_range
    cdef unsigned long bandwidth_limit
    cdef unsigned long long bytes_in
    cdef unsigned long long bytes_out
    cdef object last_frame_times
    cdef object file
    cdef object frame_types
    cdef object blank_buffer
    cdef uint64_t first_frame_timestamp
    cdef uint8_t ready
    cdef uint8_t need_reconfig

    cdef object __weakref__

    def init_context(self, encoding: str, unsigned int width, unsigned int height, src_format: str,
                     options: typedict) -> None:
        log("enc_x264.init_context%s", (encoding, width, height, src_format, options))
        global COLORSPACE_FORMATS, generation
        cs_info = COLORSPACE_FORMATS.get(src_format)
        assert cs_info is not None, "invalid source format: %s, must be one of: %s" % (src_format, COLORSPACE_FORMATS.keys())
        assert encoding=="h264", "invalid encoding: %s" % encoding
        assert options.intget("scaled-width", width)==width, "x264 encoder does not handle scaling"
        assert options.intget("scaled-height", height)==height, "x264 encoder does not handle scaling"
        if width%2 != 0 and src_format not in ("BGRX", "YUV444P"):
            raise ValueError(f"invalid odd width {width} for {src_format}")
        if height%2 != 0 and src_format=="YUV420P":
            raise ValueError(f"invalid odd height {height} for {src_format}")
        self.width = width
        self.height = height
        self.quality = options.intget("quality", 50)
        self.speed = options.intget("speed", 50)
        self.full_range = 0
        #self.opencl = USE_OPENCL and width>=32 and height>=32
        self.content_type = options.strget("content-type", "unknown")      #ie: "video"
        self.b_frames = 0 if MB_INFO else options.intget("b-frames", 0)
        self.fast_decode = options.boolget("h264.fast-decode", False)
        self.max_delayed = options.intget("max-delayed", MAX_DELAYED_FRAMES) * int(not self.fast_decode) * int(self.b_frames)
        self.preset = get_x264_preset(self.speed)
        self.src_format = src_format
        self.csc_format = COLORSPACES[src_format]
        self.colorspace = cs_info[0]
        self.frames = 0
        self.frame_types = {}
        self.last_frame_times = deque(maxlen=200)
        self.time = 0
        self.need_reconfig = 0
        self.first_frame_timestamp = 0
        self.bandwidth_limit = options.intget("bandwidth-limit", 0)
        default_profile = os.environ.get("XPRA_X264_PROFILE", "")
        self.profile = get_profile(options, csc_mode=self.src_format, default_profile=default_profile)
        self.export_nals = options.intget("h264.export-nals", 0)
        profile_options = cs_info[2]
        if self.profile and self.profile not in profile_options:
            log.warn("Warning: '%s' is not a valid profile for %s", self.profile, src_format)
            log.warn(" must be one of: %s", csv(profile_options))
            self.profile = ""
        if not self.profile:
            self.profile = cs_info[1]
            log("using default profile=%s", self.profile)
        else:
            log("using profile=%s", self.profile)
        self.init_encoder(options)
        gen = generation.increase()
        self.file = None
        if SAVE_TO_FILE is not None:
            filename = SAVE_TO_FILE+"x264-"+str(gen)+".%s" % encoding
            self.file = open(filename, "wb")
            log.info(f"saving {encoding} stream to {filename!r}")
        if BLANK_VIDEO:
            self.blank_buffer = b"\0" * (self.width * self.height * 4)
        self.ready = 1

    def is_ready(self) -> bool:
        return bool(self.ready)

    def get_tune(self) -> bytes:
        log("x264: get_tune() TUNE=%s, fast_decode=%s, content_type=%s", TUNE, self.fast_decode, self.content_type)
        if TUNE:
            return TUNE.encode()
        tunes = []
        if self.content_type.find("video")>=0:
            tunes.append(b"film")
        elif self.content_type.find("text")>=0:
            tunes.append(b"grain")
            tunes.append(b"zerolatency")
        else:
            tunes.append(b"zerolatency")
        if self.fast_decode:
            tunes.append(b"fastdecode")
        return b",".join(tunes)

    cdef void init_encoder(self, options: typedict):
        cdef x264_param_t param
        cdef const char *preset = get_preset_names()[self.preset]
        self.tune = self.get_tune()
        x264_param_default_preset(&param, preset, self.tune)
        profile = b(self.profile)
        x264_param_apply_profile(&param, profile)
        self.tune_param(&param, options)

        self.context = x264_encoder_open(&param)
        cdef int maxd = x264_encoder_maximum_delayed_frames(self.context)
        log("x264 context=%#x, %7s %4ix%-4i quality=%i, speed=%i, content_type=%s", <uintptr_t> self.context, self.src_format, self.width, self.height, self.quality, self.speed, self.content_type)
        log("x264 maximum_delayed_frames=%i", maxd)
        log("x264 params: %s", self.get_param_info(&param))
        assert self.context!=NULL,  "context initialization failed for format %s" % self.src_format

    cdef void tune_param(self, x264_param_t *param, options: typedict):
        param.i_lookahead_threads = 0
        if MIN_SLICED_THREADS_SPEED>0 and self.speed>=MIN_SLICED_THREADS_SPEED and not self.fast_decode:
            param.b_sliced_threads = 1
            param.i_threads = THREADS
        else:
            #cap i_threads since i_thread_frames will be set to i_threads
            param.i_threads = min(self.max_delayed, THREADS)
        #we never lose frames or use seeking, so no need for regular I-frames:
        param.i_keyint_max = X264_KEYINT_MAX_INFINITE
        #we don't want IDR frames either:
        param.i_keyint_min = X264_KEYINT_MAX_INFINITE
        # we can't enable 'intra-refresh' here because this triggers a warning:
        # 'intra-refresh is not compatible with open-gop'
        # if param.i_frame_reference<2:
        #    param.b_intra_refresh = 1   #intra refresh uses more bandwidth, but avoids IDR spikes
        param.b_open_gop = 1        #allow open gop
        #param.b_opencl = self.opencl
        param.i_bframe = self.b_frames
        if self.bandwidth_limit>0 and self.bandwidth_limit<=5*1000*1000:
            #CBR mode:
            param.rc.i_rc_method = X264_RC_ABR
            param.rc.i_bitrate = self.bandwidth_limit//1024
            param.rc.i_vbv_max_bitrate = 2*self.bandwidth_limit//1024
            param.rc.i_vbv_buffer_size = self.bandwidth_limit//1024
            param.rc.f_vbv_buffer_init = 1
        else:
            param.rc.i_rc_method = X264_RC_CRF
        param.rc.i_lookahead = min(param.rc.i_lookahead, self.b_frames-1)
        param.b_vfr_input = 0
        if not self.b_frames:
            param.i_sync_lookahead = 0
            param.rc.b_mb_tree = 0
        else:
            param.i_sync_lookahead = max(0, min(self.max_delayed-param.i_threads-param.rc.i_lookahead, param.i_sync_lookahead))
            #don't use TRELLIS, which uses too many delayed frames:
            if param.i_bframe_adaptive==X264_B_ADAPT_TRELLIS:
                param.i_bframe_adaptive = X264_B_ADAPT_FAST
        if self.content_type!="unknown" and self.content_type.find("video")<0:
            #specifically told this is not video,
            #so use a simple motion search:
            param.analyse.i_me_method = X264_ME_DIA
        set_f_rf(param, get_x264_quality(self.quality, self.profile))
        #client can tune these options:
        param.b_open_gop = options.boolget("h264.open-gop", param.b_open_gop)
        param.b_deblocking_filter = not self.fast_decode and options.boolget("h264.deblocking-filter", param.b_deblocking_filter)
        param.b_cabac = not self.fast_decode and options.boolget("h264.cabac", param.b_cabac)
        param.b_bluray_compat = options.boolget("h264.bluray-compat", param.b_bluray_compat)
        if self.fast_decode:
            param.analyse.b_weighted_bipred = 0
            param.analyse.i_weighted_pred = X264_WEIGHTP_NONE
        #input format:
        param.i_width = self.width
        param.i_height = self.height
        param.i_csp = self.colorspace
        if (self.colorspace & X264_CSP_HIGH_DEPTH)>0:
            param.i_bitdepth = 10
        else:
            param.i_bitdepth = 8
        param.vui.i_sar_height = 0
        param.vui.i_sar_width = 0
        param.vui.i_overscan = 1        # no overscan
        param.vui.i_vidformat = 0
        param.vui.b_fullrange = self.full_range
        param.vui.i_colorprim = 1
        param.vui.i_transfer = 1
        param.vui.i_colmatrix = 0
        param.vui.i_chroma_loc = 0

        if MB_INFO:
            param.analyse.b_fast_pskip = 1
            param.analyse.b_mb_info = 1

        #logging hook:
        param.pf_log = <void *> X264_log
        param.i_log_level = LOG_LEVEL

    def clean(self) -> None:
        log("x264 close context %#x", <uintptr_t> self.context)
        cdef x264_t *context = self.context
        if context!=NULL:
            self.context = NULL
            x264_encoder_close(context)
        self.frames = 0
        self.width = 0
        self.height = 0
        self.fast_decode = 0
        self.src_format = ""
        self.content_type = None
        self.profile = None
        self.time = 0
        self.colorspace = 0
        self.preset = 0
        self.quality = 0
        self.speed = 0
        self.bytes_in = 0
        self.bytes_out = 0
        self.last_frame_times = []
        self.first_frame_timestamp = 0
        f = self.file
        if f:
            self.file = None
            f.close()

    def get_info(self) -> Dict[str, Any]:
        cdef double pps
        if self.profile is None:
            return {}
        info = get_info()
        info |= {
            "profile"       : self.profile,
            "preset"        : get_preset_names()[self.preset],
            "fast-decode"   : bool(self.fast_decode),
            "max-delayed"   : self.max_delayed,
            "b-frames"      : self.b_frames,
            "tune"          : self.tune or "",
            "frames"        : int(self.frames),
            "width"         : self.width,
            "height"        : self.height,
            #"opencl"        : bool(self.opencl),
            "speed"         : self.speed,
            "quality"       : self.quality,
            "lossless"      : self.quality==100,
            "src_format"    : self.src_format,
            "csc_format"    : self.csc_format,
            "content-type"  : self.content_type,
            "frame-types"   : self.frame_types,
            "delayed"       : self.delayed_frames,
            "bandwidth-limit" : int(self.bandwidth_limit),
        }
        cdef x264_param_t param
        x264_encoder_parameters(self.context, &param)
        info["params"] = self.get_param_info(&param)
        if self.bytes_in>0 and self.bytes_out>0:
            info |= {
                "bytes_in"  : int(self.bytes_in),
                "bytes_out" : int(self.bytes_out),
                "ratio_pct" : int(100.0 * self.bytes_out / self.bytes_in),
            }
        if self.frames>0 and self.time>0:
            pps = self.width * self.height * self.frames / self.time
            info |= {
                "total_time_ms"     : int(self.time * 1000),
                "pixels_per_second" : int(pps),
            }
        #calculate fps:
        cdef unsigned int f = 0
        cdef double now = monotonic()
        cdef double last_time = now
        cdef double cut_off = now-10.0
        cdef double ms_per_frame = 0
        for start,end in tuple(self.last_frame_times):
            if end>cut_off:
                f += 1
                last_time = min(last_time, end)
                ms_per_frame += (end-start)
        if f>0 and last_time<now:
            info |= {
                "fps"           : int(0.5+f/(now-last_time)),
                "ms_per_frame"  : int(1000.0*ms_per_frame/f),
            }
        return info

    cdef dict get_param_info(self, x264_param_t *param):
        return {
            "me"                : self.get_analyse_info(param),
            "rc"                : self.get_rc_info(param),
            "vui"               : self.get_vui_info(param),
            "vfr-input"         : bool(param.b_vfr_input),
            "bframe-adaptive"   :  ADAPT_TYPES.get(param.i_bframe_adaptive, param.i_bframe_adaptive),
            "open-gop"          : bool(param.b_open_gop),
            "bluray-compat"     : bool(param.b_bluray_compat),
            "cabac"             : bool(param.b_cabac),
            "deblocking-filter" : bool(param.b_deblocking_filter),
            "intra-refresh"     : bool(param.b_intra_refresh),
            "interlaced"        : bool(param.b_interlaced),
            "constrained_intra" : bool(param.b_constrained_intra),
            "threads"           : {0 : "auto"}.get(param.i_threads, param.i_threads),
            "sliced-threads"    : bool(param.b_sliced_threads),
        }

    cdef dict get_vui_info(self, x264_param_t *param):
        return {
            # "overscan": bool(param.vui.i_overscan),
            "fullrange" : bool(param.vui.b_fullrange),
        }

    cdef dict get_analyse_info(self, x264_param_t *param):
        return {
            "type"              : ME_TYPES.get(param.analyse.i_me_method, param.analyse.i_me_method),
            "me-range"          : param.analyse.i_me_range,
            "mv-range"          : param.analyse.i_mv_range,
            "mv-range-thread"   : param.analyse.i_mv_range_thread,
            "subpel_refine"     : param.analyse.i_subpel_refine,
            "weighted-pred"     : param.analyse.i_weighted_pred,
            "intra"             : param.analyse.intra,
            "inter"             : param.analyse.inter,
            "fast_pskip"        : bool(param.analyse.b_fast_pskip),
            "mb_info"           : bool(param.analyse.b_mb_info),
        }

    cdef dict get_rc_info(self, x264_param_t *param):
        return {
            "rc-method"         : RC_TYPES.get(param.rc.i_rc_method, param.rc.i_rc_method),
            "qp_constant"       : param.rc.i_qp_constant,
            "qp_min"            : param.rc.i_qp_min,
            "qp_max"            : param.rc.i_qp_max,
            "qp_step"           : param.rc.i_qp_step,
            "bitrate"           : param.rc.i_bitrate,
            "vbv_max_bitrate"   : param.rc.i_vbv_max_bitrate,
            "vbv_buffer_size"   : param.rc.i_vbv_buffer_size,
            #"vbv_buffer_init"   : param.rc.f_vbv_buffer_init, #can't have floats with bencoder
            "vbv_max_bitrate"   : param.rc.i_vbv_max_bitrate,

            "mb-tree"           : bool(param.rc.b_mb_tree),
            "lookahead"         : param.rc.i_lookahead,
        }

    def __repr__(self):
        if not self.src_format:
            return "x264_encoder(uninitialized)"
        return "x264_encoder(%s - %sx%s)" % (self.src_format, self.width, self.height)

    def is_closed(self) -> bool:
        return self.context==NULL

    def get_encoding(self) -> str:
        return "h264"

    def __dealloc__(self):
        self.clean()

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_type(self) -> str:
        return  "x264"

    def get_src_format(self) -> str:
        return self.src_format

    def compress_image(self, image: ImageWrapper, options=None) -> Tuple[bytes, Dict]:
        cdef x264_picture_t pic_in
        cdef int i

        assert image.get_width()>=self.width
        assert image.get_height()>=self.height
        pf = image.get_pixel_format().replace("A", "X")
        if pf != self.src_format:
            raise ValueError(f"expected {self.src_format} but got {image.get_pixel_format()}")

        if self.first_frame_timestamp==0:
            self.first_frame_timestamp = image.get_timestamp()

        cdef int full_range = image.get_full_range()
        if full_range != self.full_range:
            self.full_range = full_range
            self.need_reconfig = 1

        options = typedict(options or {})
        content_type = options.strget("content-type", self.content_type)
        b_frames = options.intget("b-frames", 0)
        if content_type!=self.content_type or self.b_frames!=b_frames:
            #some options have changed:
            log("compress_image: reconfig b-frames=%s, content_type=%s (from %s, %s)", b_frames, content_type, self.b_frames, self.content_type)
            self.content_type = content_type
            self.b_frames = b_frames
            self.need_reconfig = 1

        cdef int speed = options.intget("speed", 50)
        if speed>=0:
            self.set_encoding_speed(speed)
        else:
            speed = self.speed
        cdef int quality = options.intget("quality", -1)
        if quality>=0:
            self.set_encoding_quality(quality)
        else:
            quality = self.quality

        if BLANK_VIDEO:
            if image.get_planes()==3:
                pixels = (self.blank_buffer, self.blank_buffer, self.blank_buffer)
            else:
                pixels = self.blank_buffer
        else:
            pixels = image.get_pixels()
        assert pixels, "failed to get pixels from %s" % image
        istrides = image.get_rowstride()

        x264_picture_init(&pic_in)
        cdef Py_buffer py_buf[3]
        for i in range(3):
            pic_in.img.plane[i] = NULL
            pic_in.img.i_stride[i] = 0
            memset(&py_buf[i], 0, sizeof(Py_buffer))

        cdef int mb_count = roundup(self.width, 16) // 16 * roundup(self.height, 16) // 16
        cdef uint8_t *mb_info = NULL
        if MB_INFO:
            mb_info = <uint8_t*> memalign(mb_count)
            memset(mb_info, 0, mb_count)
            pic_in.prop.mb_info_free = &free
            # for i in range(mb_count):
            #     mb_info[i] = i % 2
        pic_in.prop.mb_info = mb_info

        try:
            if self.src_format.find("RGB")>=0 or self.src_format.find("BGR")>=0:
                assert len(pixels)>0
                assert istrides>0
                if PyObject_GetBuffer(pixels, &py_buf[0], PyBUF_ANY_CONTIGUOUS):
                    raise ValueError(f"failed to read pixel data from {type(pixels)}")
                pic_in.img.plane[0] = <uint8_t*> py_buf.buf
                pic_in.img.i_stride[0] = istrides
                self.bytes_in += py_buf.len
                pic_in.img.i_plane = 1
            else:
                assert len(pixels)==3, "image pixels does not have 3 planes! (found %s)" % len(pixels)
                assert len(istrides)==3, "image strides does not have 3 values! (found %s)" % len(istrides)
                for i in range(3):
                    if PyObject_GetBuffer(pixels[i], &py_buf[i], PyBUF_ANY_CONTIGUOUS):
                        raise ValueError(f"failed to read pixel data from {type(pixels[i])}")
                    pic_in.img.plane[i] = <uint8_t*> py_buf[i].buf
                    pic_in.img.i_stride[i] = istrides[i]
                    self.bytes_in += py_buf[i].len
                pic_in.img.i_plane = 3
            pic_in.img.i_csp = self.colorspace
            pic_in.i_pts = image.get_timestamp()-self.first_frame_timestamp
            return self.do_compress_image(&pic_in, quality, speed)
        finally:
            for i in range(3):
                if py_buf[i].buf:
                    PyBuffer_Release(&py_buf[i])

    cdef object do_compress_image(self, x264_picture_t *pic_in, int quality=-1, int speed=-1):
        cdef x264_nal_t *nals = NULL
        cdef int i_nals = 0
        cdef x264_picture_t pic_out
        assert self.context!=NULL
        cdef double start = monotonic()

        cdef int frame_size
        with nogil:
            x264_picture_init(&pic_out)
            frame_size = x264_encoder_encode(self.context, &nals, &i_nals, pic_in, &pic_out)
        if frame_size < 0:
            log.error("x264 encoding error: frame_size is invalid!")
            return None
        self.delayed_frames = x264_encoder_delayed_frames(self.context)
        if i_nals==0:
            if self.delayed_frames>0:
                log("x264 encode %i delayed frames after %i", self.delayed_frames, self.frames)
                return b"", {
                    "delayed" : self.delayed_frames,
                    "frame"   : self.frames,
                    }
            raise RuntimeError("x264_encoder_encode produced no data (frame=%i, frame-size=%i, b-frames=%s, delayed-frames=%i)" % (self.frames, frame_size, self.b_frames, self.delayed_frames))
        slice_type = SLICE_TYPES.get(pic_out.i_type, pic_out.i_type)
        self.frame_types[slice_type] = self.frame_types.get(slice_type, 0)+1
        log("x264 encode %7s frame %5i as %4s slice with %i nals, tune=%s, total %7i bytes, keyframe=%-5s, delayed=%i",
            self.src_format, self.frames, slice_type, i_nals, s(self.tune), frame_size, bool(pic_out.b_keyframe), self.delayed_frames)
        bnals = []
        nal_indexes = []
        cdef unsigned int index = 0
        for i in range(i_nals):
            out = <char *>nals[i].p_payload
            cdata = out[:nals[i].i_payload]
            bnals.append(cdata)
            index += nals[i].i_payload
            nal_indexes.append(index)
            if LOG_NALS:
                log.info(" nal %s priority:%10s, type:%10s, payload=%#x, payload size=%i",
                         i, NAL_PRIORITIES.get(nals[i].i_ref_idc, nals[i].i_ref_idc), NAL_TYPES.get(nals[i].i_type, nals[i].i_type), <uintptr_t> nals[i].p_payload, nals[i].i_payload)
        cdata = b"".join(bnals)
        if len(cdata)!=frame_size:
            log.warn("Warning: h264 nals do not match frame size")
            log.warn(" expected %i bytes, but got %i nals and %i bytes", frame_size, len(bnals), len(cdata))
        self.bytes_out += frame_size
        #restore speed and quality if we temporarily modified them:
        if speed>=0:
            self.set_encoding_speed(self.speed)
        if quality>=0:
            self.set_encoding_quality(self.quality)
        if self.need_reconfig:
            self.reconfig_tune()
        #info for client:
        client_options = {
            "frame"     : int(self.frames),
            "pts"       : int(pic_out.i_pts),
            # "quality"   : max(0, min(100, quality)),
            # "speed"     : max(0, min(100, speed)),
            "csc"       : self.csc_format,
            "full-range" : bool(self.full_range),
        }
        if slice_type!="P":
            client_options["type"] = slice_type
        if self.delayed_frames>0:
            client_options["delayed"] = self.delayed_frames
        if self.export_nals:
            client_options["nals"] = nal_indexes
        if self.frames == 0:
            client_options["profile"] = self.profile
        #accounting:
        cdef double end = monotonic()
        self.frames += 1
        self.time += end-start
        self.last_frame_times.append((start, end))
        assert self.context!=NULL
        if self.file and frame_size>0:
            self.file.write(cdata)
            self.file.flush()
        return cdata, client_options

    def flush(self, unsigned long frame_no) -> None:
        if self.frames>frame_no or self.context==NULL:
            return None, {}
        self.delayed_frames = x264_encoder_delayed_frames(self.context)
        log("x264 flush(%i) %i delayed frames", frame_no, self.delayed_frames)
        if self.delayed_frames<=0:
            return None, {}
        cdef x264_picture_t pic_out
        x264_picture_init(&pic_out)
        return self.do_compress_image(NULL)

    def set_encoding_speed(self, int pct) -> None:
        assert pct>=0 and pct<=100, "invalid percentage: %s" % pct
        assert self.context!=NULL, "context is closed!"
        cdef x264_param_t param
        cdef int new_preset = get_x264_preset(pct)
        if new_preset == self.preset:
            return
        self.preset = new_preset
        self.speed = pct
        #retrieve current parameters:
        x264_encoder_parameters(self.context, &param)
        #apply new preset:
        self.tune = self.get_tune()
        x264_param_default_preset(&param, get_preset_names()[new_preset], self.tune)
        self.need_reconfig = 1

    def set_encoding_quality(self, int pct) -> None:
        assert pct>=0 and pct<=100, "invalid percentage: %s" % pct
        if self.quality==pct:
            return
        if abs(self.quality - pct)<=4 and pct!=100 and self.quality!=100:
            #not enough of a change to bother (but always change to/from 100)
            return
        #adjust quality:
        self.quality = pct
        self.need_reconfig = 1

    def reconfig_tune(self) -> None:
        self.need_reconfig = 0
        assert self.context!=NULL, "context is closed!"
        cdef x264_param_t param
        x264_encoder_parameters(self.context, &param)
        self.tune_param(&param, typedict())
        if x264_encoder_reconfig(self.context, &param) != 0:
            raise RuntimeError("x264_encoder_reconfig failed")


def selftest(full=False) -> None:
    log("enc_x264 selftest: %s", get_info())
    global SAVE_TO_FILE
    from xpra.codecs.checks import testencoder, get_encoder_max_sizes
    from xpra.codecs.x264 import encoder
    temp = SAVE_TO_FILE
    try:
        SAVE_TO_FILE = None
        assert testencoder(encoder, full)
        #this is expensive, so don't run it unless "full" is set:
        if full:
            global MAX_WIDTH, MAX_HEIGHT
            maxw, maxh = get_encoder_max_sizes(encoder)
            assert maxw>=MAX_WIDTH and maxh>=MAX_HEIGHT, "%s is limited to %ix%i and not %ix%i" % (encoder, maxw, maxh, MAX_WIDTH, MAX_HEIGHT)
            MAX_WIDTH, MAX_HEIGHT = maxw, maxh
            log.info("%s max dimensions: %ix%i", encoder, MAX_WIDTH, MAX_HEIGHT)
    finally:
        SAVE_TO_FILE = temp



#!/usr/bin/env python3
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.gtk.window import add_close_accel
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.os_util import gi_import

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")


class BellWindow(Gtk.Window):
    def __init__(self):
        super().__init__()
        self.set_position(Gtk.WindowPosition.CENTER)
        self.set_default_size(320, 120)
        self.set_title("Test System Bell")
        self.connect("destroy", Gtk.main_quit)
        icon = get_icon_pixbuf("bell.png")
        if icon:
            self.set_icon(icon)
        btn = Gtk.Button(label="default bell")
        btn.connect('clicked', self.bell)
        self.add(btn)

    def show_with_focus(self):
        force_focus()
        self.show_all()
        super().present()

    def bell(self, *_args):
        from xpra.platform.gui import system_bell
        system_bell(self.get_window().get_xid(), 0, 100, 2000, 1000, 0, 0, "test")


def main():
    from xpra.gtk.signals import quit_on_signals
    from xpra.gtk.util import init_display_source
    with program_context("bell", "Bell"):
        init_display_source()
        w = BellWindow()
        add_close_accel(w, Gtk.main_quit)
        GLib.idle_add(w.show_with_focus)
        quit_on_signals("bell test window")
        Gtk.main()
        return 0


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
from typing import Union
from collections.abc import Callable

from aioquic.asyncio import QuicConnectionProtocol
from aioquic.h0.connection import H0Connection
from aioquic.h3.connection import H3Connection
from aioquic.h3.events import H3Event
from aioquic.quic.packet import QuicErrorCode

from xpra.net.quic.common import SERVER_NAME, http_date, binary_headers
from xpra.net.http.directory_listing import list_directory
from xpra.net.http.handler import DIRECTORY_LISTING, translate_path, load_path, may_reload_headers
from xpra.net.common import HttpResponse
from xpra.util.str_fn import Ellipsizer
from xpra.log import Logger

log = Logger("quic")

HttpConnection = Union[H0Connection, H3Connection]


class HttpRequestHandler:
    def __init__(self, xpra_server: object,
                 authority: bytes,
                 connection: HttpConnection,
                 protocol: QuicConnectionProtocol,
                 scope: dict,
                 stream_id: int,
                 transmit: Callable[[], None],
                 ) -> None:
        self.xpra_server = xpra_server
        self.authority = authority
        self.connection = connection
        self.protocol = protocol
        self.scope = scope
        self.stream_id = stream_id
        self.transmit = transmit

    def send_http3_response(self, code, headers: dict, body: bytes = b"") -> None:
        self.send_response_header(code, headers)
        if body:
            self.send_response_body(body)
        self.transmit()

    def send_response_header(self, status: int, headers: dict) -> None:
        full_headers = {
            ":status": str(status),
            "server": SERVER_NAME.encode(),
            "date": http_date().encode(),
        }
        full_headers.update(headers)
        self.connection.send_headers(stream_id=self.stream_id, headers=binary_headers(full_headers))

    def send_response_body(self, body: bytes = b"", more_body: bool = False) -> None:
        self.connection.send_data(stream_id=self.stream_id, data=body, end_stream=not more_body)

    def http_event_received(self, event: H3Event) -> None:
        log(f"http_event_received(%s) scope={self.scope}", Ellipsizer(event))
        http_version = self.scope.get("http_version", "0")
        if http_version != "3":
            message = "http version {http_version} is not supported"
            log.error(f"Error: {message}")
            self.protocol.close(QuicErrorCode.APPLICATION_ERROR, message)
            return
        method = self.scope.get("method", "")
        req_path = self.scope.get("path", "")
        log.info(f"HTTP request {method} {req_path}")
        scripts = self.xpra_server.get_http_scripts()
        # script is a: Callable[[str], HttpResponse]
        script = scripts.get(req_path)
        log(f"req_path={req_path}, scripts={scripts}")
        if script:
            log(f"request for {req_path} handled using {script}")
            script_response: HttpResponse = script(req_path)
            self.send_http3_response(*script_response)
            return
        if method != "GET":
            message = f"http {method} requests are not supported"
            log.warn(f"Warning: {message}")
            self.protocol.close(QuicErrorCode.APPLICATION_ERROR, message)
            return
        self.handle_get_request(req_path)

    def handle_get_request(self, req_path: str) -> None:
        web_root = self.xpra_server._www_dir
        headers_dirs = self.xpra_server._http_headers_dirs
        headers = may_reload_headers(headers_dirs)
        log(f"handle_get_request({req_path}) web_root={web_root}, headers_dir={headers_dirs}")
        path = translate_path(req_path, web_root)
        if not path or not os.path.exists(path):
            self.send_http3_response(404, headers, body=b"Path not found")
            return
        if os.path.isdir(path):
            if not path.endswith('/'):
                # redirect browser - doing basically what apache does
                headers["Location"] = path + "/"
                self.send_http3_response(301, headers)
                return
            for index in "index.html", "index.htm":
                index = os.path.join(path, index)
                if os.path.exists(index):
                    path = index
                    break
            else:
                if not DIRECTORY_LISTING:
                    self.send_http3_response(403, headers, body=b"Directory listing forbidden")
                    return
                code, extra_headers, body = list_directory(path)
                headers.update(extra_headers)
                self.send_http3_response(code, headers, body)
                return
        accept_encoding = self.scope.get("headers", {}).get("accept-encoding", "").split(",")
        code, path_headers, body = load_path(accept_encoding, path)
        headers.update(path_headers)
        self.send_http3_response(code, headers, body)



#!/usr/bin/env python3

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GLib	#pylint: disable=wrong-import-position @UnresolvedImport

def main():
	window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	window.set_size_request(400, 200)
	window.connect("delete_event", Gtk.main_quit)
	window.iconify()
	GLib.timeout_add(2000, window.deiconify)
	Gtk.main()
	return 0


if __name__ == "__main__":
	main()



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.client.gtk3.window_base import GTKClientWindowBase, HAS_X11_BINDINGS
from xpra.gtk.widget import scaled_image
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.scripts.config import FALSE_OPTIONS
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.os_util import gi_import
from xpra.util.str_fn import bytestostr
from xpra.log import Logger

paintlog = Logger("paint")
metalog = Logger("metadata")
geomlog = Logger("geometry")

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GdkPixbuf = gi_import("GdkPixbuf")
GObject = gi_import("GObject")
Gio = gi_import("Gio")

WINDOW_ICON = envbool("XPRA_WINDOW_ICON", True)
WINDOW_XPRA_MENU = envbool("XPRA_WINDOW_XPRA_MENU", True)
WINDOW_MENU = envbool("XPRA_WINDOW_MENU", True)


class ClientWindow(GTKClientWindowBase):
    """
    GTK3 version of the ClientWindow class
    """

    def init_window(self, metadata: typedict) -> None:
        super().init_window(metadata)
        self.menu_helper = None
        self.header_bar_image = None
        if self.can_use_header_bar(metadata):
            self.add_header_bar()

    def _icon_size(self) -> int:
        tb = self.get_titlebar()
        try:
            h = tb.get_preferred_size()[-1] - 8
        except Exception:
            h = 24
        return min(128, max(h, 24))

    def set_icon(self, pixbuf: GdkPixbuf.Pixbuf) -> None:
        super().set_icon(pixbuf)
        hbi = self.header_bar_image
        if hbi and WINDOW_ICON:
            h = self._icon_size()
            pixbuf = pixbuf.scale_simple(h, h, GdkPixbuf.InterpType.HYPER)
            hbi.set_from_pixbuf(pixbuf)

    def can_use_header_bar(self, metadata: typedict) -> bool:
        if self.is_OR() or not self.get_decorated():
            return False
        hbl = (self.headerbar or "").lower().strip()
        if hbl in FALSE_OPTIONS:
            return False
        if hbl == "force":
            return True
        # we can't enable it if there are size-constraints:
        sc = metadata.dictget("size-constraints")
        if sc is None:
            return True
        tsc = typedict(sc)
        maxs = tsc.intpair("maximum-size")
        if maxs:
            return False
        mins = tsc.intpair("minimum-size")
        if mins and mins != (0, 0):
            return False
        if tsc.intpair("increment", (0, 0)) != (0, 0):
            return False
        return True

    def add_header_bar(self) -> None:
        metalog("add_header_bar()")
        hb = Gtk.HeaderBar()
        hb.set_has_subtitle(False)
        hb.set_show_close_button(True)
        hb.props.title = self.get_title()
        if WINDOW_MENU:
            # the icon 'open-menu-symbolic' will be replaced with the window icon
            # when we receive it
            icon = Gio.ThemedIcon(name="preferences-system-windows")
            self.header_bar_image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
            button = Gtk.Button()
            button.add(self.header_bar_image)
            button.connect("clicked", self.show_window_menu)
            hb.pack_start(button)
        elif WINDOW_ICON:
            # just the icon, no menu:
            pixbuf = get_icon_pixbuf("transparent.png")
            self.header_bar_image = scaled_image(pixbuf, self._icon_size())
            hb.pack_start(self.header_bar_image)
        if WINDOW_XPRA_MENU:
            icon = Gio.ThemedIcon(name="open-menu-symbolic")
            image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
            button = Gtk.Button()
            button.add(image)
            button.connect("clicked", self.show_xpra_menu)
            hb.pack_end(button)
        self.set_titlebar(hb)

    def show_xpra_menu(self, *_args) -> None:
        mh = self._client.get_menu_helper()
        if mh:
            mh.build()
            mh.popup(0, 0)

    def show_window_menu(self, *_args) -> None:
        if not self.menu_helper:
            from xpra.client.gtk3.window_menu import WindowMenuHelper
            self.menu_helper = WindowMenuHelper(self._client, self)
            self.menu_helper.build()
        self.menu_helper.popup(0, 0)

    def get_backing_class(self) -> type:
        from xpra.client.gtk3.cairo_backing import CairoBacking
        return CairoBacking

    def xget_u32_property(self, target, name: str):
        if HAS_X11_BINDINGS:
            return GTKClientWindowBase.xget_u32_property(self, target, name)
        # pure Gdk lookup:
        try:
            name_atom = Gdk.Atom.intern(name, False)
            type_atom = Gdk.Atom.intern("CARDINAL", False)
            prop = Gdk.property_get(target, name_atom, type_atom, 0, 9999, False)
            if not prop or len(prop) != 3 or len(prop[2]) != 1:
                return None
            metalog("xget_u32_property(%s, %s)=%s", target, name, prop[2][0])
            return prop[2][0]
        except Exception as e:
            metalog.error("xget_u32_property error on %s / %s: %s", target, name, e)

    def get_drawing_area_geometry(self) -> tuple[int, int, int, int]:
        gdkwindow = self.drawing_area.get_window()
        if gdkwindow:
            x, y = gdkwindow.get_origin()[1:]
        else:
            x, y = self.get_position()
        w, h = self.get_size()
        return x, y, w, h

    def apply_geometry_hints(self, hints: typedict) -> None:
        """ we convert the hints as a dict into a gdk.Geometry + gdk.WindowHints """
        wh = Gdk.WindowHints
        name_to_hint = {
            "max_width": wh.MAX_SIZE,
            "max_height": wh.MAX_SIZE,
            "min_width": wh.MIN_SIZE,
            "min_height": wh.MIN_SIZE,
            "base_width": wh.BASE_SIZE,
            "base_height": wh.BASE_SIZE,
            "width_inc": wh.RESIZE_INC,
            "height_inc": wh.RESIZE_INC,
            "min_aspect_ratio": wh.ASPECT,
            "max_aspect_ratio": wh.ASPECT,
        }
        # these fields can be copied directly to the gdk.Geometry as ints:
        INT_FIELDS: list[str] = [
            "min_width", "min_height",
            "max_width", "max_height",
            "base_width", "base_height",
            "width_inc", "height_inc",
        ]
        ASPECT_FIELDS: dict[str, str] = {
            "min_aspect_ratio": "min_aspect",
            "max_aspect_ratio": "max_aspect",
        }
        thints = typedict(hints)
        if self.drawing_area:
            # apply min size to the drawing_area:
            # (for CSD mode, ie: headerbar)
            minw = thints.intget("min_width", 0)
            minh = thints.intget("min_height", 0)
            self.drawing_area.set_size_request(minw, minh)

        geom = Gdk.Geometry()
        mask = 0
        for k, v in hints.items():
            k = bytestostr(k)
            if k in INT_FIELDS:
                setattr(geom, k, v)
                mask |= int(name_to_hint.get(k, 0))
            elif k in ASPECT_FIELDS:
                field = ASPECT_FIELDS[k]
                setattr(geom, field, float(v))
                mask |= int(name_to_hint.get(k, 0))
        gdk_hints = Gdk.WindowHints(mask)
        geomlog("apply_geometry_hints(%s) geometry=%s, hints=%s", hints, geom, gdk_hints)
        self.set_geometry_hints(self.drawing_area, geom, gdk_hints)

    def can_maximize(self) -> bool:
        hints = self.geometry_hints
        if not hints:
            return True
        maxw = hints.intget("max_width", 32768)
        maxh = hints.intget("max_height", 32768)
        if maxw > 32000 and maxh > 32000:
            return True
        geom = self.get_drawing_area_geometry()
        dw, dh = geom[2], geom[3]
        return dw < maxw and dh < maxh

    def draw_widget(self, widget, context) -> bool:
        paintlog("draw_widget(%s, %s)", widget, context)
        if not self.get_mapped():
            return False
        backing = self._backing
        if not backing:
            return False
        context.save()
        self.paint_backing_offset_border(backing, context)
        self.clip_to_backing(backing, context)
        backing.cairo_draw(context)
        context.restore()
        self.cairo_paint_border(context, None)
        if not self._client.server_ok():
            self.paint_spinner(context)
        return True


GObject.type_register(ClientWindow)



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False

from time import monotonic
from typing import Any, Dict, Tuple
from collections.abc import Sequence

from xpra.codecs.constants import VideoSpec
from xpra.util.objects import typedict
from xpra.codecs.image import ImageWrapper
from xpra.log import Logger
log = Logger("encoder", "openh264")

from libcpp cimport bool as bool_t
from libc.string cimport memset
from libc.stdint cimport uint8_t, uintptr_t
from xpra.buffers.membuf cimport buffer_context  # pylint: disable=syntax-error

cdef extern from "wels/codec_app_def.h":
    ctypedef enum DECODING_STATE:
        dsErrorFree             # bit stream error-free
        dsFramePending          # need more throughput to generate a frame output,
        dsRefLost               # layer lost at reference frame with temporal id 0
        dsBitstreamError        # error bitstreams(maybe broken internal frame) the decoder cared
        dsDepLayerLost          # dependented layer is ever lost
        dsNoParamSets           # no parameter set NALs involved
        dsDataErrorConcealed    # current data error concealed specified
        dsRefListNullPtrs       # ref picure list contains null ptrs within uiRefCount range

        # Errors derived from logic level
        dsInvalidArgument       # invalid argument specified
        dsInitialOptExpected    # initializing operation is expected
        dsOutOfMemory           # out of memory due to new request
        dsDstBufNeedExpan       # actual picture size exceeds size of dst pBuffer feed in decoder, so need expand its size

    ctypedef struct OpenH264Version:
        unsigned int uMajor
        unsigned int uMinor
        unsigned int uRevision
        unsigned int uReserved

cdef extern from "wels/codec_def.h":
    ctypedef struct SSysMEMBuffer:
        int iWidth                          #width of decoded pic for display
        int iHeight                         #height of decoded pic for display
        int iFormat                         #type is "EVideoFormatType"
        int iStride[2]                      #stride of 2 component

    cdef union UsrData:
        SSysMEMBuffer sSystemBuffer         #memory info for one picture

    ctypedef struct SBufferInfo:
        int iBufferStatus                   #0: one frame data is not ready; 1: one frame data is ready
        unsigned long long uiInBsTimeStamp  #input BS timestamp
        unsigned long long uiOutYuvTimeStamp#output YUV timestamp, when bufferstatus is 1
        UsrData UsrData #output buffer info
        unsigned char* pDst[3]              #point to picture YUV data

cdef extern from "wels/codec_api.h":
    void WelsGetCodecVersionEx(OpenH264Version* pVersion)
    long WelsCreateDecoder(ISVCDecoder** ppDecoder)
    void WelsDestroyDecoder(ISVCDecoder* pDecoder)

    ctypedef enum ERROR_CON_IDC:
        ERROR_CON_DISABLE
        ERROR_CON_FRAME_COPY
        ERROR_CON_SLICE_COPY
        ERROR_CON_FRAME_COPY_CROSS_IDR
        ERROR_CON_SLICE_COPY_CROSS_IDR
        ERROR_CON_SLICE_COPY_CROSS_IDR_FREEZE_RES_CHANGE
        ERROR_CON_SLICE_MV_COPY_CROSS_IDR
        ERROR_CON_SLICE_MV_COPY_CROSS_IDR_FREEZE_RES_CHANGE

    ctypedef enum VIDEO_BITSTREAM_TYPE:
        VIDEO_BITSTREAM_AVC
        VIDEO_BITSTREAM_SVC
        VIDEO_BITSTREAM_DEFAULT

    ctypedef struct SVideoProperty:
        unsigned int          size          # size of the struct
        VIDEO_BITSTREAM_TYPE  eVideoBsType

    ctypedef struct SDecodingParam:
        char*     pFileNameRestructed       # file name of reconstructed frame used for PSNR calculation based debug
        unsigned int  uiCpuLoad             # CPU load
        unsigned char uiTargetDqLayer       # setting target dq layer id
        ERROR_CON_IDC eEcActiveIdc          # whether active error concealment feature in decoder
        bool_t bParseOnly                     # decoder for parse only, no reconstruction. When it is true, SPS/PPS size should not exceed SPS_PPS_BS_SIZE (128). Otherwise, it will return error info
        SVideoProperty   sVideoProperty

    cdef cppclass ISVCDecoder:
        long Initialize(const SDecodingParam* pParam)
        long Uninitialize()
        long DecodeFrameNoDelay(const unsigned char* pSrc, const int iSrcLen,
                                unsigned char** ppDst, SBufferInfo* pDstInfo) nogil

ERROR_STR: Dict[int, str] = {
    dsErrorFree: "no error",
    dsFramePending: "need more throughput to generate a frame output",
    dsRefLost: "layer lost at reference frame with temporal id 0",
    dsBitstreamError: "error bitstreams(maybe broken internal frame) the decoder cared",
    dsDepLayerLost: "dependented layer is ever lost",
    dsNoParamSets: "no parameter set NALs involved",
    dsDataErrorConcealed: "current data error concealed specified",
    dsRefListNullPtrs: "ref picure list contains null ptrs within uiRefCount range",

    # Errors derived from logic level
    dsInvalidArgument: "invalid argument specified",
    dsInitialOptExpected: "initializing operation is expected",
    dsOutOfMemory: "out of memory due to new request",
    dsDstBufNeedExpan: "actual picture size exceeds size of dst pBuffer feed in decoder, so need expand its size",
}


def get_version() -> Tuple[int, int, int]:
    cdef OpenH264Version version
    WelsGetCodecVersionEx(&version)
    return (version.uMajor, version.uMinor, version.uRevision)


def get_type() -> str:
    return "openh264"


def get_info() -> Dict[str, Any]:
    return {
        "version"   : get_version(),
    }


def get_encodings() -> Sequence[str]:
    return ("h264", )


def get_min_size(encoding) -> Tuple[int, int]:
    return 32, 32


MAX_WIDTH, MAX_HEIGHT = (8192, 4096)


def get_specs() -> Sequence[VideoSpec]:
    return (
        VideoSpec(
            encoding="h264", input_colorspace="YUV420P", output_colorspaces=("YUV420P", ),
            has_lossless_mode=False,
            codec_class=Decoder, codec_type=get_type(),
            quality=40, speed=20,
            size_efficiency=40,
            setup_cost=0, width_mask=0xFFFE, height_mask=0xFFFE,
            max_w=MAX_WIDTH, max_h=MAX_HEIGHT,
        ),
    )


cdef class Decoder:
    cdef ISVCDecoder *context;
    cdef unsigned long frames
    cdef int width
    cdef int height
    cdef object colorspace

    cdef object __weakref__

    def init_context(self, encoding: str, int width, int height, colorspace: str, options: typedict) -> None:
        log("openh264.init_context%s", (encoding, width, height, colorspace))
        assert encoding == "h264", f"invalid encoding: {encoding}"
        assert colorspace == "YUV420P", f"invalid colorspace: {colorspace}"
        self.width = width
        self.height = height
        self.colorspace = colorspace
        self.frames = 0
        cdef long r = WelsCreateDecoder(&self.context)
        if r:
            raise RuntimeError(f"error {r} creating openh264 decoder")
        cdef SDecodingParam dec_param
        memset(&dec_param, 0, sizeof(SDecodingParam))
        #dec_param.sVideoProperty.eVideoBsType = VIDEO_BITSTREAM_AVC;
        #for Parsing only, the assignment is mandatory
        #dec_param.bParseOnly = 0
        self.context.Initialize(&dec_param)

    def get_encoding(self) -> str:
        return "h264"

    def get_colorspace(self) -> str:
        return self.colorspace

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def is_closed(self) -> bool:
        return self.context!=NULL

    def get_type(self) -> str:
        return "openh264"

    def __dealloc__(self):
        self.clean()

    def clean(self) -> None:
        log("openh264 close context %#x", <uintptr_t> self.context)
        cdef ISVCDecoder *context = self.context
        if context:
            context.Uninitialize()
            self.context = NULL
            WelsDestroyDecoder(context)
        self.frames = 0
        self.width = 0
        self.height = 0
        self.colorspace = ""

    def get_info(self) -> Dict[str, Any]:
        info = get_info()
        info |= {
            "frames"        : int(self.frames),
            "width"         : self.width,
            "height"        : self.height,
            "colorspace"    : self.colorspace,
        }
        return info

    def decompress_image(self, data: bytes, options: typedict) -> ImageWrapper:
        cdef SBufferInfo buf_info
        cdef long r = 0
        cdef unsigned char* src
        cdef int src_len = 0
        cdef uint8_t *yuv[3]
        start = monotonic()
        with buffer_context(data) as bc:
            src = <unsigned char*> (<uintptr_t> int(bc))
            src_len = len(bc)
            with nogil:
                r = self.context.DecodeFrameNoDelay(<const unsigned char*> src, <const int> src_len, yuv, &buf_info)
        if r:
            msg = ERROR_STR.get(r, "unknown")
            raise RuntimeError(f"openh264 frame decoding error {msg!r}")
        end = monotonic()
        cdef int ystride = buf_info.UsrData.sSystemBuffer.iStride[0]
        cdef int uvstride = buf_info.UsrData.sSystemBuffer.iStride[1]
        strides = (ystride, uvstride, uvstride)
        cdef int width = buf_info.UsrData.sSystemBuffer.iWidth
        cdef int height = buf_info.UsrData.sSystemBuffer.iHeight
        log(f"openh264 strides=%s, size=%s", strides, (width, height))
        if width < self.width:
            raise ValueError(f"stream width {width} is smaller than decoder width {self.width}")
        if height < self.height:
            raise ValueError(f"stream width {height} is smaller than decoder width {self.height}")
        cdef int wdelta = width - self.width
        cdef int hdelta = height - self.height
        if abs(wdelta) > 1 or abs(hdelta) > 1:
            if (wdelta & 0xffe0) > 0 or (hdelta & 0xfffe0) > 0:
                log.warn("Warning: image bigger than expected")
                log.warn(f" {width}x{height} instead of {self.width}x{self.height}")
        pixels = (
            yuv[0][:ystride*height],
            yuv[1][:uvstride*(height//2)],
            yuv[2][:uvstride*(height//2)],
        )
        log(f"openh264 decoded {src_len:8} bytes into {width}x{height} YUV420P in {int((end-start)*1000):3}ms")
        full_range = options.boolget("full-range")
        return ImageWrapper(0, 0, self.width, self.height, pixels, self.colorspace, 24, strides, 1, ImageWrapper.PLANAR_3, full_range=full_range)


def selftest(full=False) -> None:
    log("openh264 selftest: %s", get_info())
    from xpra.codecs.checks import testdecoder
    from xpra.codecs.openh264 import decoder
    testdecoder(decoder, full)



#!/usr/bin/env python3
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.gtk.util import get_default_root_window
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import label
from xpra.gtk.pixbuf import get_icon_pixbuf

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")


def make_window():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_title("Window Transient Test")
    window.set_size_request(400, 300)
    window.set_position(Gtk.WindowPosition.CENTER)
    window.connect("delete_event", Gtk.main_quit)
    icon = get_icon_pixbuf("windows.png")
    if icon:
        window.set_icon(icon)
    vbox = Gtk.VBox(homogeneous=False, spacing=0)

    btn = Gtk.Button(label="Create Transient")

    def create_transient(*_args):
        tw = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
        tw.set_size_request(200, 100)
        if icon:
            tw.set_icon(icon)
        tw.connect("delete_event", lambda x, y: tw.destroy())
        tw.set_transient_for(window)
        tw.add(label("Transient Window"))
        tw.show_all()

    btn.connect('clicked', create_transient)
    vbox.pack_start(btn, expand=False, fill=False, padding=10)

    btn = Gtk.Button(label="Create Transient (with 5 second delay)")

    def delayed_transient(*_args):
        GLib.timeout_add(5000, create_transient)

    btn.connect('clicked', delayed_transient)
    vbox.pack_start(btn, expand=False, fill=False, padding=10)

    btn = Gtk.Button(label="Create Root Transient")

    def create_root_transient(*_args):
        tw = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
        tw.set_size_request(200, 100)
        if icon:
            tw.set_icon(icon)
        tw.connect("delete_event", lambda x, y: tw.destroy())
        tw.realize()
        tw.get_window().set_transient_for(get_default_root_window())
        tw.add(label("Transient Root Window"))
        tw.show_all()

    btn.connect('clicked', create_root_transient)
    vbox.pack_start(btn, expand=False, fill=False, padding=10)
    window.add(vbox)
    return window


def main():
    with program_context("window-transient", "Window Transient"):
        w = make_window()
        add_close_accel(w, Gtk.main_quit)
        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("transient window test")

        def show_with_focus():
            force_focus()
            w.show_all()
            w.present()

        GLib.idle_add(show_with_focus)
        Gtk.main()
        return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.platform import (
    program_context, threaded_server_init,
    get_application_name, get_prgname,
    )


class PlatformInitTest(unittest.TestCase):

    def test_all(self):
        threaded_server_init()
        with program_context() as f:
            assert repr(f)

        #this doesn't actually exercise more code,
        #as the attributes are initialized once per process...
        with program_context("prg") as f:
            assert repr(f)

        with program_context(None, "app") as f:
            assert repr(f)

        with program_context("prg", "app") as f:
            assert repr(f)
            assert get_application_name()=="app"
            assert get_prgname()=="prg"


def main():
    unittest.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Callable, Sequence, Sized
from dataclasses import dataclass

from xpra.util.env import envbool
from xpra.util.str_fn import memoryview_to_bytes
from xpra.common import MIN_COMPRESS_SIZE, MAX_DECOMPRESSED_SIZE, SizedBuffer

# all the compressors we know about:
ALL_COMPRESSORS: Sequence[str] = ("lz4", "zlib", "lzo", "brotli", "none")

VALID_COMPRESSORS: Sequence[str] = ("lz4", "brotli")

# the compressors we may want to use, in the best compatibility order:
TRY_COMPRESSORS: Sequence[str] = ("lz4", "brotli", "none")
# order for performance:
PERFORMANCE_ORDER: Sequence[str] = ("none", "lz4", "brotli")
# require compression (disallow 'none'):
PERFORMANCE_COMPRESSION: Sequence[str] = ("lz4", "brotli")


@dataclass
class Compression:
    name: str
    version: str
    compress: Callable[[SizedBuffer, int], tuple[int, SizedBuffer]]
    decompress: Callable[[SizedBuffer], SizedBuffer]


COMPRESSION: dict[str, Compression] = {}


def init_lz4() -> Compression:
    # pylint: disable=import-outside-toplevel
    # pylint: disable=redefined-outer-name
    from xpra.net.lz4.lz4 import compress, decompress, get_version  # @UnresolvedImport
    from xpra.net.protocol.header import LZ4_FLAG

    def lz4_compress(data: SizedBuffer, level: int) -> tuple[int, memoryview]:
        flag = min(15, level) | LZ4_FLAG
        return flag, compress(data, acceleration=max(0, 5 - level // 3))

    def lz4_decompress(data: SizedBuffer) -> memoryview:
        return decompress(data, max_size=MAX_DECOMPRESSED_SIZE)

    return Compression("lz4", get_version(), lz4_compress, lz4_decompress)


def init_brotli() -> Compression:
    # pylint: disable=import-outside-toplevel
    # pylint: disable=redefined-outer-name
    from xpra.net.protocol.header import BROTLI_FLAG
    from xpra.net.brotli.compressor import compress, get_version  # @UnresolvedImport
    from xpra.net.brotli.decompressor import decompress  # @UnresolvedImport
    brotli_decompress = decompress
    brotli_compress = compress
    brotli_version = get_version()

    def brotli_compress_shim(packet: SizedBuffer, level: int) -> tuple[int, memoryview]:
        if len(packet) > 1024 * 1024:
            level = min(9, level)
        else:
            level = min(11, level)
        if not isinstance(packet, (bytes, bytearray, memoryview)):
            packet = bytes(str(packet), 'UTF-8')
        return level | BROTLI_FLAG, brotli_compress(packet, quality=level)

    return Compression("brotli", brotli_version, brotli_compress_shim, brotli_decompress)


def init_none() -> Compression:

    def nocompress(data: SizedBuffer, _level) -> tuple[int, SizedBuffer]:
        if isinstance(data, bytes):
            return 0, data
        return 0, memoryview_to_bytes(data)

    def nodecompress(v: SizedBuffer) -> SizedBuffer:
        return v

    return Compression("none", "0", nocompress, nodecompress)


def init_compressors(*names: str) -> None:
    for x in names:
        assert x not in ("compressors", "all"), "attempted to recurse!"
        if not envbool("XPRA_" + x.upper(), True):
            continue
        attr = globals().get(f"init_{x}", None)
        if attr is None:
            from xpra.log import Logger
            logger = Logger("network", "protocol")
            logger.warn(f"Warning: invalid compressor {x} specified")
            continue
        try:
            if not callable(attr):
                raise ValueError(f"{attr!r} for {x} is not callable")
            fn: Callable = attr
            c = fn()
            assert c
            COMPRESSION[x] = c
        except (TypeError, ImportError, AttributeError):
            # pylint: disable=import-outside-toplevel
            from xpra.log import Logger
            logger = Logger("network", "protocol")
            logger(f"no {x}", exc_info=True)


def init_all() -> None:
    init_compressors(*(list(TRY_COMPRESSORS) + ["none"]))


def use(compressor) -> bool:
    return compressor in COMPRESSION


def get_compression_caps(full_info: int = 1) -> dict[str, Any]:
    caps: dict[str, Any] = {}
    for x in TRY_COMPRESSORS:
        c = COMPRESSION.get(x)
        if c is None:
            continue
        ccaps = caps.setdefault(x, {})
        if full_info > 1 and c.version:
            ccaps["version"] = c.version
        ccaps[""] = True
    return caps


def get_enabled_compressors(order=TRY_COMPRESSORS) -> Sequence[str]:
    return tuple(x for x in order if x in COMPRESSION)


def get_compressor(name: str) -> Callable:
    c = COMPRESSION.get(name)
    if c is not None:
        return c.compress
    raise ValueError(f"{name!r} compression is not supported")


class Compressed:
    __slots__ = ("datatype", "data", "can_inline")

    def __init__(self, datatype: str, data: SizedBuffer | Sequence, can_inline=True):
        if not data and not isinstance(data, Sequence):
            raise ValueError(f"missing compressed data, got {data!r} ({type(data)})")
        self.datatype = datatype
        self.data = data
        self.can_inline = can_inline

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return f"Compressed({self.datatype}: {len(self.data)} bytes)"


class LevelCompressed(Compressed):
    __slots__ = ("level", "algorithm")

    def __init__(self, datatype: str, data: SizedBuffer, level: int, algo: str, can_inline: bool):
        super().__init__(datatype, data, can_inline)
        self.level = level
        self.algorithm = algo

    def __repr__(self):
        return f"LevelCompressed({self.datatype}: {len(self.data)} bytes as {self.algorithm}/{self.level}"


class LargeStructure:
    __slots__ = ("datatype", "data")

    def __init__(self, datatype: str, data: Sized):
        self.datatype = datatype
        self.data = data

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return f"LargeStructure({self.datatype}: {len(self.data)} bytes)"


class Compressible(LargeStructure):
    __slots__ = ()

    # wrapper for data that should be compressed at some point,
    # to use this class, you must override compress()

    def __repr__(self):
        return f"Compressible({self.datatype}: {len(self.data)} bytes)"

    def compress(self):
        raise NotImplementedError(f"compress() function is not defined on {self} ({type(self)})")


def compressed_wrapper(datatype, data, level=5, can_inline=True, **kwargs) -> Compressed:
    size = len(data)

    def no() -> Compressed:
        return Compressed(f"raw {datatype}", data, can_inline=can_inline)

    if size <= MIN_COMPRESS_SIZE:
        # don't bother
        return no()
    if size > MAX_DECOMPRESSED_SIZE:
        sizemb = size // 1024 // 1024
        maxmb = MAX_DECOMPRESSED_SIZE // 1024 // 1024
        raise ValueError(f"uncompressed data is too large: {sizemb}MB, limit is {maxmb}MB")
    try:
        algo = next(x for x in PERFORMANCE_COMPRESSION if kwargs.get(x) and x in COMPRESSION)
    except StopIteration:
        return no()
        # raise InvalidCompressionException("no compressors available")
    # should use a smarter selection of algo based on datatype
    # ie: 'text' -> brotli
    c = COMPRESSION[algo]
    cl, cdata = c.compress(data, level)
    min_saving = int(kwargs.get("min_saving", 0))
    if len(cdata) >= size + min_saving:
        return no()
    return LevelCompressed(datatype, cdata, cl, algo, can_inline=can_inline)


class InvalidCompressionException(Exception):
    pass


def get_compression_type(level: int) -> str:
    from xpra.net.protocol.header import LZ4_FLAG, BROTLI_FLAG
    if level & LZ4_FLAG:
        return "lz4"
    if level & BROTLI_FLAG:
        return "brotli"
    return "zlib"


def decompress(data: bytes, level: int) -> SizedBuffer:
    from xpra.net.protocol.header import LZ4_FLAG, BROTLI_FLAG
    if level & LZ4_FLAG:
        algo = "lz4"
    elif level & BROTLI_FLAG:
        algo = "brotli"
    else:
        algo = "zlib"
    return decompress_by_name(data, algo)


def decompress_by_name(data: bytes, algo: str) -> SizedBuffer:
    c = COMPRESSION.get(algo)
    if c is None:
        raise InvalidCompressionException(f"{algo} is not available")
    return c.decompress(data)


def main():  # pragma: no cover
    # pylint: disable=import-outside-toplevel
    from xpra.util.str_fn import print_nested_dict
    from xpra.platform import program_context
    with program_context("Compression", "Compression Info"):
        init_all()
        print_nested_dict(get_compression_caps())


if __name__ == "__main__":  # pragma: no cover
    main()



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.common import MAX_WINDOW_SIZE
from xpra.log import Logger

log = Logger("x11", "window")

MAX_ASPECT = 2 ** 15 - 1


def sanitize_size_hints(size_hints: dict[str, Any]) -> None:
    """
        Some applications may set nonsensical values,
        try our best to come up with something that can actually be used.
    """
    if not size_hints:
        return
    for attr in ("min-aspect", "max-aspect"):
        v = size_hints.get(attr)
        if v is not None:
            try:
                f = float(v)
            except ValueError:
                f = None
            if f is None or f <= 0 or f >= MAX_ASPECT:
                log.warn("Warning: clearing invalid aspect hint value for %s: %s", attr, v)
                del size_hints[attr]
    for attr in ("minimum-aspect-ratio", "maximum-aspect-ratio"):
        v = size_hints.get(attr)
        if v is not None:
            if v[1] == 0:
                f = None
            else:
                try:
                    f = float(v[0]) / float(v[1])
                except (ValueError, ZeroDivisionError):
                    f = None
            if f is None or f <= 0 or f >= MAX_ASPECT:
                log.warn("Warning: clearing invalid aspect hint value for %s: %s", attr, v)
                del size_hints[attr]
    for attr in ("maximum-size", "minimum-size", "base-size", "increment"):
        v = size_hints.get(attr)
        if v is not None:
            try:
                w, h = v
                int(w)
                int(h)
            except (ValueError, TypeError):
                w, h = None, None
            if (w is None or h is None) or w >= MAX_WINDOW_SIZE or h >= MAX_WINDOW_SIZE:
                log("clearing invalid size hint value for %s: %s", attr, v)
                del size_hints[attr]
    # if max-size is smaller than min-size (bogus), clamp it..
    clamped = False
    mins = size_hints.get("minimum-size")
    if mins:
        minw, minh = mins
        if minw <= 0 and minh <= 0:
            # doesn't do anything
            size_hints.pop("minimum-size")
            mins = None
            clamped = True
    maxs = size_hints.get("maximum-size")
    if maxs:
        maxw, maxh = maxs
        if maxw <= 0 and maxh <= 0:
            # doesn't make sense!
            size_hints["maximum-size"] = None
            maxs = None
            clamped = True
    if mins and maxs:
        minw, minh = mins
        maxw, maxh = maxs
        if minw > 0 and minw > maxw:
            # min higher than max!
            if minw <= 256:
                maxw = minw
            elif maxw >= 256:
                minw = maxw
            else:
                minw = maxw = 256
            clamped = True
        if minh > 0 and minh > maxh:
            # min higher than max!
            if minh <= 256:
                maxh = minh
            elif maxh >= 256:
                minh = maxh
            else:
                minh = maxh = 256
            clamped = True
        if clamped:
            size_hints["minimum-size"] = minw, minh
            size_hints["maximum-size"] = maxw, maxh
    if clamped:
        log.warn("Warning: invalid size hints")
        log.warn(" min_size=%s / max_size=%s", mins, maxs)
        log.warn(" changed to: %s / %s", size_hints.get("minimum-size"), size_hints.get("maximum-size"))



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from math import sqrt
from time import monotonic
from typing import Deque, Any

from collections import deque
from xpra.util.stats import get_list_stats, get_weighted_list_stats
from xpra.util.str_fn import csv
from xpra.util.env import envint
from xpra.server.cystats import (
    logp,
    calculate_time_weighted_average,
    calculate_size_weighted_average,
    calculate_timesize_weighted_average,
    calculate_for_average,
)

from xpra.log import Logger

log = Logger("stats")

# how many historical records to keep
# for the various statistics we collect:
# (cannot be lower than DamageBatchConfig.max_events)
NRECS: int = 100

TARGET_LATENCY_TOLERANCE = envint("XPRA_TARGET_LATENCY_TOLERANCE", 20) / 1000.0


class WindowPerformanceStatistics:
    """
    Statistics which belong to a specific WindowSource
    """

    def __init__(self):
        self.reset()

    # assume 100ms until we get some data to compute the real values
    DEFAULT_DAMAGE_LATENCY: float = 0.1
    DEFAULT_NETWORK_LATENCY: float = 0.1
    DEFAULT_TARGET_LATENCY: float = 0.1

    def reset(self) -> None:
        self.init_time: float = monotonic()
        # records how long it took the client to decode frames:
        # (ack_time, no of pixels, decoding_time*1000*1000)
        self.client_decode_time: Deque[tuple[float, int, int]] = deque(maxlen=NRECS)
        # encoding: (time, coding, pixels, bpp, compressed_size, encoding_time)
        self.encoding_stats: Deque[tuple[float, str, int, int, int, float]] = deque(maxlen=NRECS)
        # records how long it took for a damage request to be sent
        # last NRECS: (sent_time, no of pixels, actual batch delay, damage_latency)
        self.damage_in_latency: Deque[tuple[float, int, float, float]] = deque(maxlen=NRECS)
        # records how long it took for a damage request to be processed
        # last NRECS: (processed_time, no of pixels, actual batch delay, damage_latency)
        self.damage_out_latency: Deque[tuple[float, int, float, float]] = deque(maxlen=NRECS)
        # records when damage packets are sent,
        # so we can calculate the "client_latency" when the client sends
        # the corresponding ack ("damage-sequence" packet - see "client_ack_damage")
        self.damage_ack_pending: dict[int, tuple[float, str, int, int, dict, float]] = {}
        # for each encoding, how many frames we sent and how many pixels in total
        self.encoding_totals: dict[str, list[int]] = {}
        # damage regions waiting to be picked up by the encoding thread:
        # for each sequence no: (damage_time, w, h)
        self.encoding_pending: dict = {}
        # every time we get a damage event, we record: time,x,y,w,h
        self.last_damage_events: Deque[tuple[float, int, int, int, int]] = deque(maxlen=4 * NRECS)
        self.last_damage_event_time: float = 0
        self.last_recalculate: float = 0
        self.damage_events_count = 0
        self.packet_count = 0

        self.resize_events: Deque[float] = deque(maxlen=4)  # (time)
        self.last_resized: float = 0
        self.last_packet_time: float = 0

        # these values are calculated from the values above (see update_averages)
        self.target_latency = self.DEFAULT_TARGET_LATENCY
        self.avg_damage_in_latency = self.DEFAULT_DAMAGE_LATENCY
        self.recent_damage_in_latency = self.DEFAULT_DAMAGE_LATENCY
        self.avg_damage_out_latency = self.DEFAULT_DAMAGE_LATENCY + self.DEFAULT_NETWORK_LATENCY
        self.recent_damage_out_latency = self.DEFAULT_DAMAGE_LATENCY + self.DEFAULT_NETWORK_LATENCY
        self.max_latency = self.DEFAULT_DAMAGE_LATENCY + self.DEFAULT_NETWORK_LATENCY
        self.avg_decode_speed = -1
        self.recent_decode_speed = -1

    def reset_backlog(self) -> None:
        # this should be a last resort..
        self.damage_ack_pending = {}

    def update_averages(self) -> None:
        # damage "in" latency: (the time it takes for damage requests to be processed only)
        dil = tuple(self.damage_in_latency)
        if dil:
            data = tuple((when, latency) for when, _, _, latency in dil)
            self.avg_damage_in_latency, self.recent_damage_in_latency = calculate_time_weighted_average(data)
        # damage "out" latency: (the time it takes for damage requests to be processed and sent out)
        dol = tuple(self.damage_out_latency)
        if dol:
            data = tuple((when, latency) for when, _, _, latency in dol)
            self.avg_damage_out_latency, self.recent_damage_out_latency = calculate_time_weighted_average(data)
        # client decode speed:
        cdt = tuple(self.client_decode_time)
        if cdt:
            # the elapsed time recorded is in microseconds:
            decode_speed = tuple(
                (event_time, size, int(size * 1000 * 1000 / elapsed))
                for event_time, size, elapsed in cdt if elapsed > 0
            )
            r = calculate_size_weighted_average(decode_speed)
            self.avg_decode_speed = int(r[0])
            self.recent_decode_speed = int(r[1])
        # network send speed:
        all_l = [0.1,
                 self.avg_damage_in_latency, self.recent_damage_in_latency,
                 self.avg_damage_out_latency, self.recent_damage_out_latency]
        self.max_latency = max(all_l)

    def get_factors(self, bandwidth_limit=0) -> list[tuple[str, str, float, float]]:
        factors = []

        def mayaddfac(metric, info, factor, weight):
            if weight > 0.01:
                factors.append((metric, info, factor, weight))

        # ratio of "in" and "out" latency indicates network bottleneck:
        # (the difference between the two is the time it takes to send)
        if self.damage_in_latency and self.damage_out_latency:
            # prevent jitter from skewing the values too much
            ad = max(0.010, 0.040 + self.avg_damage_out_latency - self.avg_damage_in_latency)
            rd = max(0.010, 0.040 + self.recent_damage_out_latency - self.recent_damage_in_latency)
            metric = "damage-network-delay"
            # info: avg delay=%.3f recent delay=%.3f" % (ad, rd)
            mayaddfac(*calculate_for_average(metric, ad, rd))
        # client decode time:
        ads = self.avg_decode_speed
        rds = self.recent_decode_speed
        if ads > 0 and rds > 0:
            metric = "client-decode-speed"
            # info: avg=%.1f, recent=%.1f (MPixels/s)" % (ads/1000/1000, self.recent_decode_speed/1000/1000)
            # our calculate methods aims for lower values, so invert speed
            # this is how long it takes to send 1MB:
            avg1MB = 1.0 * 1024 * 1024 / ads
            recent1MB = 1.0 * 1024 * 1024 / rds
            weight_div = max(0.25, rds / (4 * 1000 * 1000))
            mayaddfac(*calculate_for_average(metric, avg1MB, recent1MB, weight_offset=0.0, weight_div=weight_div))
        ldet = self.last_damage_event_time
        if ldet:
            # If nothing happens for a while then we can reduce the batch delay,
            # however we must ensure this is not caused by a high system latency,
            # so we ignore short elapsed times.
            elapsed = monotonic() - ldet
            mtime = max(0.0, elapsed - self.max_latency * 2)
            # the longer the time, the more we slash:
            weight = sqrt(mtime)
            target = max(0.0, 1.0 - mtime)
            metric = "damage-rate"
            info = {"elapsed": int(1000.0 * elapsed),
                    "max_latency": int(1000.0 * self.max_latency)}
            mayaddfac(metric, info, target, weight)
        if bandwidth_limit > 0:
            # calculate how much bandwidth we have used in the last second (in bps):
            # encoding_stats.append((end, coding, w*h, bpp, len(data), end-start))
            cutoff = monotonic() - 1
            used = sum(v[4] for v in tuple(self.encoding_stats) if v[0] > cutoff) * 8
            info = {
                "budget": bandwidth_limit,
                "used": used,
            }
            # aim for 10% below the limit:
            target = used * 110.0 / 100.0 / bandwidth_limit
            # if we are getting close to or above the limit,
            # the certainty of this factor goes up:
            weight = max(0.0, target - 1) * (5 + logp(target))
            mayaddfac("bandwidth-limit", info, target, weight)
        return factors

    def get_info(self) -> dict[str, Any]:
        info = {
            "damage": {
                "events": self.damage_events_count,
                "packets_sent": self.packet_count,
                "target-latency": int(1000 * self.target_latency),
            }
        }
        # encoding stats:
        estats = tuple(self.encoding_stats)
        if estats:
            def add_compression_stats(enc_stats, encoding=None):
                comp_ratios_pct = []
                comp_times_ns = []
                total_pixels = 0
                total_time = 0.0
                for _, _, pixels, bpp, compressed_size, compression_time in enc_stats:
                    if compressed_size > 0 and pixels > 0:
                        osize = pixels * bpp / 8
                        comp_ratios_pct.append((100.0 * compressed_size / osize, pixels))
                        comp_times_ns.append((1000.0 * 1000 * 1000 * compression_time / pixels, pixels))
                        total_pixels += pixels
                        total_time += compression_time
                einfo: dict[str, Any] = info.setdefault(encoding or "encoding", {})
                einfo["ratio_pct"] = get_weighted_list_stats(comp_ratios_pct)
                einfo["pixels_per_ns"] = get_weighted_list_stats(comp_times_ns)
                if total_time > 0:
                    einfo["pixels_encoded_per_second"] = int(total_pixels / total_time)

            add_compression_stats(estats)
            encodings_used = tuple(x[1] for x in estats)
            for encoding in encodings_used:
                enc_stats = tuple(x for x in estats if x[1] == encoding)
                add_compression_stats(enc_stats, encoding)

        dinfo = info.setdefault("damage", {})
        latencies = tuple(x[-1] * 1000 for x in tuple(self.damage_in_latency))
        dinfo["in_latency"] = get_list_stats(latencies, show_percentile=(9,))
        latencies = tuple(x[-1] * 1000 for x in tuple(self.damage_out_latency))
        dinfo["out_latency"] = get_list_stats(latencies, show_percentile=(9,))
        # per encoding totals:
        if self.encoding_totals:
            tf = info.setdefault("total_frames", {})
            tp = info.setdefault("total_pixels", {})
            for encoding, totals in tuple(self.encoding_totals.items()):
                tf[encoding] = totals[0]
                tp[encoding] = totals[1]
        return info

    def get_target_client_latency(self, min_client_latency, avg_client_latency, abs_min=0.010, jitter=0) -> float:
        """ geometric mean of the minimum (+20%) and average latency
            but not higher than twice more than the minimum,
            and not lower than abs_min.
            Then we add the average decoding latency.
            """
        decoding_latency = 0.010
        cdt = tuple(self.client_decode_time)
        if cdt:
            decoding_latency = calculate_timesize_weighted_average(cdt)[0] / 1000.0
        min_latency = max(abs_min, min_client_latency or abs_min) * 1.2
        avg_latency = max(min_latency, avg_client_latency or abs_min)
        max_latency = min(avg_latency, 4.0 * min_latency + 0.100)
        return max(abs_min, min(max_latency, sqrt(min_latency * avg_latency))) + decoding_latency + jitter / 1000.0

    def get_client_backlog(self) -> tuple[int, int]:
        packets_backlog, pixels_backlog = 0, 0
        if self.damage_ack_pending:
            queued_before = monotonic() - (self.target_latency + TARGET_LATENCY_TOLERANCE)
            dropped_acks_time = monotonic() - 60  # 1 minute
            drop_missing_acks = []
            for sequence, item in tuple(self.damage_ack_pending.items()):
                queued_at = item[0]
                if queued_at > queued_before:
                    continue
                if queued_at < dropped_acks_time:
                    drop_missing_acks.append(sequence)
                else:
                    # ack_pending = (now, coding, pixcount, bytecount, client_options, damage_time)
                    pixels = item[2]
                    packets_backlog += 1
                    pixels_backlog += pixels
            log("get_client_backlog missing acks: %s", drop_missing_acks)
            # this should never happen...
            if drop_missing_acks:
                log.error("Error: expiring %i missing damage ACKs,", len(drop_missing_acks))
                log.error(" connection may be closed or closing,")
                log.error(" sequence numbers missing: %s", csv(drop_missing_acks))
                for sequence in drop_missing_acks:
                    self.damage_ack_pending.pop(sequence, None)
        return packets_backlog, pixels_backlog

    def get_acks_pending(self) -> int:
        return sum(1 for x in self.damage_ack_pending.values() if x[0] != 0)

    def get_late_acks(self, latency) -> int:
        now = monotonic()
        sent_before = now - latency
        late = sum(1 for item in self.damage_ack_pending.values() if item[0] <= sent_before)
        log("get_late_acks(%i)=%i (%i in full pending list)", 1000 * latency, late, len(self.damage_ack_pending))
        return late

    def get_pixels_encoding_backlog(self) -> tuple[int, int]:
        pixels, count = 0, 0
        for _, w, h in self.encoding_pending.values():
            pixels += w * h
            count += 1
        return pixels, count

    def get_bitrate(self, max_elapsed: float = 1) -> int:
        cutoff = monotonic() - max_elapsed
        recs = tuple((v[0], v[4]) for v in tuple(self.encoding_stats) if v[0] >= cutoff)
        if len(recs) < 2:
            return 0
        bits = sum(v[1] for v in recs) * 8
        elapsed = recs[-1][0] - recs[0][0]
        if elapsed == 0:
            return 0
        return int(bits / elapsed)

    def get_damage_pixels(self, elapsed: float = 1) -> int:
        cutoff = monotonic() - elapsed
        return sum(v[3] * v[4] for v in tuple(self.last_damage_events) if v[0] > cutoff)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from xpra.os_util import get_hex_uuid
from xpra.util.env import osexpand
from unit.server_test_util import ServerTestUtil, log

class X11ClientTestUtil(ServerTestUtil):

	uq = 0

	def run_client(self, *args):
		client_display = self.find_free_display()
		xvfb = self.start_Xvfb(client_display)
		xvfb.display = client_display
		return xvfb, self.do_run_client(client_display, *args)

	def do_run_client(self, client_display, *args):
		from xpra.x11.vfb_util import xauth_add
		xauth_data = get_hex_uuid()
		xauthority = self.default_env.get("XAUTHORITY", osexpand("~/.Xauthority"))
		xauth_add(xauthority, client_display, xauth_data, os.getuid(), os.getgid())
		env = self.get_run_env()
		env["DISPLAY"] = client_display
		env["XPRA_LOG_PREFIX"] = "client %i: " % X11ClientTestUtil.uq
		X11ClientTestUtil.uq +=1
		log("starting test client on Xvfb %s", client_display)
		return self.run_xpra(["attach"] + list(args) , env=env)
